"""
Module containing the spacing functions for the binarycpython package. Very under-populated at the moment, but more are likely to come soon

Tasks:
    TODO: add more spacing functions to this module.
"""

from typing import Union
import functools
import math
import numpy as np
import py_rinterpolate
import sys
from binarycpython.utils.grid import Population


@functools.lru_cache(maxsize=16)
def const(
    min_bound: Union[int, float], max_bound: Union[int, float], steps: int
) -> list:
    """
    Samples a range linearly. Uses numpy linspace.

    Args:
        min_bound: lower bound of range
        max_bound: upper bound of range
        steps: number of segments between min_bound and max_bound

    Returns:
        np.linspace(min_bound, max_bound, steps+1)
    """
    return np.linspace(min_bound, max_bound, steps + 1)


############################################################
@functools.lru_cache(maxsize=16)
def const_ranges(ranges) -> list:
    """
    Samples a series of ranges linearly.

    Args:
        ranges: a tuple of tuples passed to the const() spacing function.

    Returns:
        numpy array of masses

    Example:
        The following allocates 10 stars between 0.1 and 0.65, 20 stars between 0.65
        and 0.85, and 10 stars between 0.85 and 10.0 Msun.

        samplerfunc="const_ranges((({},{},{}),({},{},{}),({},{},{})))".format(
            0.1,0.65,10,
            0.65,0.85,20,
            0.85,10.0,10
        ),

    """

    masses = np.empty(0)
    for range in ranges:
        masses = np.append(masses, const(*range))
    return np.unique(masses)


############################################################
def peak_normalized_gaussian_func(
    x: Union[int, float], mean: Union[int, float], sigma: Union[int, float]
) -> Union[int, float]:
    """
    Function to evaluate a Gaussian at a given point, note
    that the normalization is such that the peak is always 1.0,
    not that the integral is 1.0

    Args:
        x: location at which to evaluate the distribution
        mean: mean of the Gaussian
        sigma: standard deviation of the Gaussian

    Returns:
        value of the Gaussian at x
    """
    gaussian_prefactor = 1.0  # / math.sqrt(2.0 * math.pi)

    r = 1.0 / sigma
    y = (x - mean) * r
    return math.exp(-0.5 * y ** 2)


############################################################
@functools.lru_cache(maxsize=16)
def gaussian_zoom(
    min_bound: Union[int, float],
    max_bound: Union[int, float],
    zoom_mean: Union[int, float],
    zoom_dispersion: Union[int, float],
    zoom_magnitude: Union[int, float],
    steps: int,
) -> list:
    """
    Samples such that a region is zoomed in according to a 1-Gaussian function

    Args:
        min_bound: lower bound of range
        max_bound: upper bound of range
        zoom_mean: mean of the Gaussian zoom location
        zoom_dispersion: dispersion of the Gaussian
        zoom_magnitude: depth of the Gaussian (should be 0<= zoom_magntiude <1)
        steps: number of segments between min_bound and max_bound assuming a linear step
               this is what you'd normally call "resolution"

    Returns:
        Numpy array of sample values
    """

    # linear spacing: this is what we'd have
    # in the absence of a Gaussian zoom
    linear_spacing = (max_bound - min_bound) / (steps - 1)

    # make the list of values
    x = min_bound
    array = np.array([])
    while x <= max_bound:
        array = np.append(array, x)
        g = peak_normalized_gaussian_func(x, zoom_mean, zoom_dispersion)
        f = 1.0 - zoom_magnitude * g
        dx = linear_spacing * f
        x = x + dx

    # force the last array member to be max_bound if it's not
    if array[-1] != max_bound:
        array[-1] = max_bound

    return np.unique(array)


@functools.lru_cache(maxsize=16)
def const_dt(
    self,
    dt=1000.0,
    dlogt=0.1,
    mmin=0.07,
    mmax=100.0,
    nres=1000,
    logspacing=False,
    tmin=3.0,  # start at 3Myr
    tmax=None,  # use max_evolution_time by default
    mindm=None,  # tuple of tuples
    maxdm=((0.07, 1.0, 0.1), (1.0, 300.0, 1.0)),  # tuple of tuples
    fsample=1.0,
    factor=1.0,
    logmasses=False,
    log10masses=False,
    showlist=False,
    showtable=False,
):
    """
    const_dt returns a list of masses spaced at a constant age difference

    Args:
        dt: the time difference between the masses (1000.0 Myr, used when logspacing==False)
        dlogt : the delta log10(time) difference between masses (0.1 dex, used when logspacing==True)
        mmin: the minimum mass to be considered in the stellar lifetime interpolation table (0.07 Msun)
        mmax: the maximum mass to be considered in the stellar lifetime interpolation table (100.0 Msun)
        nres: the resolution of the stellar lifetime interpolation table (100)
        logspacing: whether to use log-spaced time, in which case dt is actually d(log10(t))
        tmin: the minimum time to consider (Myr, default 3.0 Myr)
        tmax: the maximum time to consider (Myr, default None which means we use the grid option 'max_evolution_time')
        mindm: a tuple of tuples containing a mass range and minimum mass spacing in that range. The default is ((0.07,1.0,0.1),(1.0,300.0,1.0)) allocated a minimum dm of 0.1Msun in the mass range 0.07 to 1.0 Msun and 1.0Msun in the range 1.0 to 300.0 Msun. Anything you set overrides this. Note, if you use only one tuple, you must set it with a trailing comma, thus, e.g. ((0.07,1.0,0.1),). (default None)
        maxdm: a list of tuples similar to mindm but specifying a maximum mass spacing. In the case of maxdm, if the third option in each tuple is negative it is treated as a log step (its absolute value is used as the step).  (default None)
        fsample: a global sampling (Shannon-like) factor (<1) to improve resolution (default 1.0, set to smaller to improve resolution)
        factor: all masses generated are multiplied by this after generation
        showtable: if True, the mass list and times are shown to stdout after generation
        showlist: if True, show the mass list once generated
        logmasses: if True, the masses are logged with math.log()
        log10masses: if True, the masses are logged with math.log10()

    Returns:
        Array of masses.

    Example:
    # these are lines set as options to Population.add_grid_value(...)

    # linear time bins of 1Gyr
    samplerfunc="const_dt(self,dt=1000,nres=100,mmin=0.07,mmax=2.0,showtable=True)"

    # logarithmic spacing in time, generally suitable for Galactic
    # chemical evolution yield grids.
    samplerfunc="const_dt(self,dlogt=0.1,nres=100,mmin=0.07,mmax=80.0,maxdm=((0.07,1.0,0.1),(1.0,10.0,1.0),(10.0,80.0,2.0)),showtable=True,logspacing=True,fsample=1.0/4.0)"

    """

    # first, make a stellar lifetime table
    #
    # we should use the bse_options from self
    # so our lifetime_population uses the same physics
    lifetime_population = Population()
    lifetime_population.bse_options = dict(self.bse_options)

    # we only want to evolve the star during nuclear burning,
    # we don't want a dry run of the grid
    # we want to use the right number of CPU cores
    lifetime_population.set(
        do_dry_run=False,
        num_cores=self.grid_options["num_cores"],
        max_stellar_type_1=10,
        save_ensemble_chunks=False,
    )

    # make a grid in M1
    lifetime_population.add_grid_variable(
        name="lnM_1",
        parameter_name="M_1",
        longname="log Primary mass",  # == single-star mass
        valuerange=[math.log(mmin), math.log(mmax)],
        samplerfunc="const(math.log({mmin}),math.log({mmax}),{nres})".format(
            mmin=mmin, mmax=mmax, nres=nres
        ),
        probdist="1",  # dprob/dm1 : we don't care, so just set it to 1
        dphasevol="dlnM_1",
        precode="M_1=math.exp(lnM_1)",
        condition="",  # Impose a condition on this grid variable. Mostly for a check for yourself
        gridtype="edge",
    )

    # set up the parse function
    def _parse_function(self, output):
        if output:
            for line in output.splitlines():
                data = line.split()
                if data[0] == "SINGLE_STAR_LIFETIME":
                    # append (log10(mass), log10(lifetime)) tuples
                    logm = math.log10(float(data[1]))
                    logt = math.log10(float(data[2]))
                    # print(line)
                    # print("logM=",logm,"M=",10.0**logm," -> logt=",logt)
                    self.grid_results["interpolation table m->t"][logm] = logt
                    self.grid_results["interpolation table t->m"][logt] = logm

    lifetime_population.set(
        parse_function=_parse_function,
    )

    # run to build the interpolation table
    print("Running population to make lifetime interpolation table, please wait")
    lifetime_population.evolve()
    # print("Data table",lifetime_population.grid_results['interpolation table t->m'])

    # convert to nested lists for the interpolator
    #
    # make time -> mass table
    data_table_time_mass = []
    times = sorted(lifetime_population.grid_results["interpolation table t->m"].keys())
    for time in times:
        mass = lifetime_population.grid_results["interpolation table t->m"][time]
        # we have to make sure the time is monotonic (not guaranteed at high mass)
        if len(data_table_time_mass) == 0:
            data_table_time_mass.append([time, mass])
        elif mass < data_table_time_mass[-1][1]:
            data_table_time_mass.append([time, mass])

    # make mass -> time table
    data_table_mass_time = []
    masses = sorted(lifetime_population.grid_results["interpolation table m->t"].keys())
    for mass in masses:
        time = lifetime_population.grid_results["interpolation table m->t"][mass]
        data_table_mass_time.append([mass, time])

    # set up interpolators
    interpolator_time_mass = py_rinterpolate.Rinterpolate(
        table=data_table_time_mass, nparams=1, ndata=1, verbosity=0  # mass  # lifetime
    )
    interpolator_mass_time = py_rinterpolate.Rinterpolate(
        table=data_table_mass_time, nparams=1, ndata=1, verbosity=0  # lifetime  # mass
    )

    # function to get a mass given a time (Myr)
    def _mass_from_time(linear_time):
        return 10.0 ** interpolator_time_mass.interpolate([math.log10(linear_time)])[0]

    # function to get a time given a mass (Msun)
    def _time_from_mass(mass):
        return 10.0 ** interpolator_mass_time.interpolate([math.log10(mass)])[0]

    # return a unique list
    def _uniq(_list):
        return sorted(list(set(_list)))

    # format a whole list like %g
    def _format(_list):
        return [float("{x:g}".format(x=x)) for x in _list]

    # construct mass list, always include the min and max
    mass_list = [mmin, mmax]

    # first, make sure the stars are separated by only
    # maxdm
    if maxdm:
        for x in maxdm:
            range_min = x[0]
            range_max = x[1]
            dm = x[2]
            if dm < 0.0:
                # use log scale
                dlogm = -dm
                logm = math.log(mmin)
                logmmax = math.log(mmax)
                logrange_min = math.log(range_min)
                logrange_max = math.log(range_max)
                while logm <= logmmax:
                    if logm >= logrange_min and logm <= logrange_max:
                        mass_list.append(math.exp(logm))
                    logm += dlogm
            else:
                # use linear scale
                m = mmin
                while m <= mmax:
                    if m >= range_min and m <= range_max:
                        mass_list.append(m)
                    m += dm

    # start time loop at tmax or max_evolution_time
    t = tmax if tmax else self.bse_options["max_evolution_time"]

    # set default mass list
    if logspacing:
        logt = math.log10(t)
        logtmin = math.log10(tmin)
        while logt > logtmin:
            m = _mass_from_time(10.0 ** logt)
            mass_list.append(m)
            logt = max(logtmin, logt - dlogt * fsample)
    else:
        while t > tmin:
            m = _mass_from_time(t)
            mass_list.append(m)
            t = max(tmin, t - dt * fsample)

    # make mass list unique
    mass_list = _uniq(mass_list)

    if mindm:
        for x in mindm:
            range_min = x[0]
            range_max = x[1]
            mindm = x[2]
            # impose a minimum dm: if two masses in the list
            # are separated by < this, remove the second
            for index, mass in enumerate(mass_list):
                if index > 0 and mass >= range_min and mass <= range_max:
                    dm = mass_list[index] - mass_list[index - 1]
                    if dm < mindm:
                        mass_list[index - 1] = 0.0
            mass_list = _uniq(mass_list)
            if mass_list[0] == 0.0:
                mass_list.remove(0.0)

    # apply multiplication factor if given
    if factor and factor != 1.0:
        mass_list = [m * factor for m in mass_list]

    # reformat numbers
    mass_list = _format(mass_list)

    # show the mass<>time table?
    if showtable:
        twas = 0.0
        logtwas = 0.0
        for i, m in enumerate(mass_list):
            t = _time_from_mass(m)
            logt = math.log10(t)
            if twas > 0.0:
                print(
                    "{i:4d} m={m:13g} t={t:13g} log10(t)={logt:13g} dt={dt:13g} dlog10(t)={dlogt:13g}".format(
                        i=i, m=m, t=t, logt=logt, dt=twas - t, dlogt=logtwas - logt
                    )
                )
            else:
                print(
                    "{i:4d} m={m:13g} t={t:13g} log10(t)={logt:13g}".format(
                        i=i, m=m, t=t, logt=logt
                    )
                )
            twas = t
            logtwas = logt
        exit()

    # return the mass list as a numpy array
    mass_list = np.unique(np.array(mass_list))

    # perhaps log the masses
    if logmasses:
        mass_list = np.log(mass_list)
    if log10masses:
        mass_list = np.log10(mass_list)

    if showlist:
        print("const_dt mass list ({} masses)\n".format(len(mass_list)), mass_list)

    return mass_list



"""
Module containing functions that binary_c-python uses to modify
dictionaries.
"""
import copy
import collections
import astropy.units as u
import numpy as np
from collections import (
    OrderedDict,
)

# we need to convert keys to floats:
# this is ~ a factor 10 faster than David's
# recursive_change_key_to_float routine,
# probably because this version only does
# the float conversion, nothing else.
def keys_to_floats(json_data):
    # assumes nested dicts ...
    # new_data = {}

    # but this copies the variable type, but has some
    # pointless copying
    # new_data = copy.copy(json_data)
    # new_data.clear()

    # this adopts the type correctly *and* is fast
    new_data = type(json_data)()

    for k, v in json_data.items():
        if isinstance(v, list):
            v = [
                keys_to_floats(item)
                if isinstance(item, collections.abc.Mapping)
                else item
                for item in v
            ]
        elif isinstance(v, collections.abc.Mapping):
            # dict, ordereddict, etc.
            v = keys_to_floats(v)
        try:
            f = float(k)
            new_data[f] = json_data[k]
        except:
            new_data[k] = v
    return new_data


def recursive_change_key_to_float(input_dict):
    """
    Function to recursively change the key to float

    This only works if the dict contains just sub-dicts or numbers/strings.
    Does not work with lists as values
    """

    new_dict = collections.OrderedDict()  # TODO: check if this still works

    for key in input_dict:
        if isinstance(input_dict[key], (dict, collections.OrderedDict)):
            try:
                num_key = float(key)
                new_dict[num_key] = recursive_change_key_to_float(
                    copy.deepcopy(input_dict[key])
                )
            except ValueError:
                new_dict[key] = recursive_change_key_to_float(
                    copy.deepcopy(input_dict[key])
                )
        else:
            try:
                num_key = float(key)
                new_dict[num_key] = input_dict[key]
            except ValueError:
                new_dict[key] = input_dict[key]

    return new_dict


def recursive_change_key_to_string(input_dict):
    """
    Function to recursively change the key back to a string but this time in a format that we decide
    """

    new_dict = collections.OrderedDict()  # TODO: check if this still works

    for key in input_dict:
        if isinstance(input_dict[key], (dict, collections.OrderedDict)):
            if isinstance(key, (int, float)):
                string_key = "{:g}".format(key)
                new_dict[string_key] = recursive_change_key_to_string(
                    copy.deepcopy(input_dict[key])
                )
            else:
                new_dict[key] = recursive_change_key_to_string(
                    copy.deepcopy(input_dict[key])
                )
        else:
            if isinstance(key, (int, float)):
                string_key = "{:g}".format(key)
                new_dict[string_key] = input_dict[key]
            else:
                new_dict[key] = input_dict[key]

    return new_dict


############################################################
# code to walk a dictionary recursively based on
# https://stackoverflow.com/questions/13687924/setting-a-value-in-a-nested-python-dictionary-given-a-list-of-indices-and-value
def _nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


def _nested_get(dic, keys):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    return dic[keys[-1]]


# function to walk through the dictionary, multiplying
# only float values by a const
def _recursive_normalize_floats(path, d, const, parent=None, ignore=None):
    if not parent:
        parent = d
    for k, v in d.items():
        if ignore and k in ignore:
            continue

        if isinstance(v, float):
            path.append(k)
            # must be a float, multiply by the constant
            _nested_set(parent, path, v * const)
            path.pop()
        elif isinstance(v, str) or isinstance(v, int):
            path.append(k)
            # do nothing to strings or ints
            path.pop()
        elif v is None:
            path.append(k)
            path.pop()
        # dicts
        # note: isinstance isn't enough, we need to check the Mapping
        elif isinstance(v, collections.abc.Mapping):
            path.append(k)
            # nested dict
            _recursive_normalize_floats(path, v, const, parent=parent)
            path.pop()
        else:
            print(
                "###Type {} not recognized: {}.{}={}".format(
                    type(v), ".".join(path), k, v
                )
            )


def multiply_float_values(d, const, ignore=None):
    """
    multiply_float_values : A function to recursively multiply values of a (nested) dictionary that are floats by a constant. Nested dictionaries call this function recursively.

    Args:
        d = the dictionary
        const = the constant that multiplies float values
    """
    path = []
    _recursive_normalize_floats(path, d, const, parent=d, ignore=ignore)


def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
    """
    Function to subtract two dictionaries.

    Only allows values to be either a dict or a numerical type

    For the overlapping keys (key name present in both dicts):
        When the keys are of the same type:
            - If the types are of numerical type: subtract the value at dict 2 from dict 1.
            - If the types are both dictionaries: call this function with the subdicts

        When the keys are not of the same type:
            - if the keys are all of numerical types

    For the unique keys:
        - if the key is from dict 1: adds the value to the new dict (be it numerical value or dict)
        - If the key is from dict 2: Adds the negative of its value in case of numerical type.
            if the type is a dict, the result of subtract_dicts({}, dict_2[key]) will be set

    If the result is 0, the key will be removed from the resulting dict.
    If that results in an empty dict, the dict will be removed too.

    Args:
        dict_1: first dictionary
        dict_2: second dictionary

    Returns:
        Subtracted dictionary
    """

    # Set up new dict
    new_dict = {}

    # Define allowed numerical types
    ALLOWED_NUMERICAL_TYPES = (float, int, np.float64)

    #
    keys_1 = dict_1.keys()
    keys_2 = dict_2.keys()

    # Find overlapping keys of both dicts
    overlapping_keys = set(keys_1).intersection(set(keys_2))

    # Find the keys that are unique
    unique_to_dict_1 = set(keys_1).difference(set(keys_2))
    unique_to_dict_2 = set(keys_2).difference(set(keys_1))

    # Add the unique keys to the new dict
    for key in unique_to_dict_1:
        # If these items are numerical types
        if isinstance(dict_1[key], ALLOWED_NUMERICAL_TYPES):
            new_dict[key] = dict_1[key]
            if new_dict[key] == 0:
                del new_dict[key]

        # Else, to be safe we should deepcopy them
        elif isinstance(dict_1[key], dict):
            copy_dict = copy.deepcopy(dict_1[key])
            new_dict[key] = copy_dict
        else:
            msg = "Error: using unsupported type for key {}: {}".format(
                key, type(dict_1[key])
            )
            print(msg)
            raise ValueError(msg)

    # Add the unique keys to the new dict
    for key in unique_to_dict_2:
        # If these items are numerical type, we should add the negative of the value
        if isinstance(dict_2[key], ALLOWED_NUMERICAL_TYPES):
            new_dict[key] = -dict_2[key]
            if new_dict[key] == 0:
                del new_dict[key]

        # Else we should place the negative of that dictionary in the new place
        elif isinstance(dict_2[key], dict):
            new_dict[key] = subtract_dicts({}, dict_2[key])
        else:
            msg = "Error: using unsupported type for key {}: {}".format(
                key, type(dict_2[key])
            )
            print(msg)
            raise ValueError(msg)

    # Go over the common keys:
    for key in overlapping_keys:

        # See whether the types are actually the same
        if not type(dict_1[key]) is type(dict_2[key]):
            # Exceptions:
            if (type(dict_1[key]) in ALLOWED_NUMERICAL_TYPES) and (
                type(dict_2[key]) in ALLOWED_NUMERICAL_TYPES
            ):
                # We can safely subtract the values since they are all numeric
                new_dict[key] = dict_1[key] - dict_2[key]
                if new_dict[key] == 0:
                    del new_dict[key]

            else:
                msg = "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
                    key,
                    dict_1[key],
                    type(dict_1[key]),
                    key,
                    dict_2[key],
                    type(dict_2[key]),
                )

                print(msg)
                raise ValueError(msg)

        # This is where the keys are the same
        else:
            # If these items are numeric types
            if isinstance(dict_1[key], ALLOWED_NUMERICAL_TYPES):
                new_dict[key] = dict_1[key] - dict_2[key]

                # Remove entry if the value is 0
                if new_dict[key] == 0:
                    del new_dict[key]

            # Else, to be safe we should deepcopy them
            elif isinstance(dict_1[key], dict):
                new_dict[key] = subtract_dicts(dict_1[key], dict_2[key])

                # Remove entry if it results in an empty dict
                # TODO: write test to prevent empty dicts from showing up
                if not new_dict[key]:
                    del new_dict[key]
            else:
                msg = "Error: using unsupported type for key {}: {}".format(
                    key, type(dict_2[key])
                )
                print(msg)
                raise ValueError(msg)

    #
    return new_dict


class AutoVivificationDict(dict):
    """
    Implementation of perl's autovivification feature, by overriding the
    get item and the __iadd__ operator (https://docs.python.org/3/reference/datamodel.html?highlight=iadd#object.__iadd__)

    This allows to set values within a subdict that might not exist yet:

    Example:
        newdict = {}
        newdict['example']['mass'] += 10
        print(newdict)
        >>> {'example': {'mass': 10}}
    """

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

    def __iadd__(self, other):
        # if a value does not exist, assume it is 0.0
        try:
            self += other
        except:
            self = other
        return self


def inspect_dict(
    input_dict: dict, indent: int = 0, print_structure: bool = True
) -> dict:
    """
    Function to (recursively) inspect a (nested) dictionary.
    The object that is returned is a dictionary containing the key of the input_dict, but as value
    it will return the type of what the value would be in the input_dict

    In this way we inspect the structure of these dictionaries, rather than the exact contents.

    Args:
        input_dict: dictionary you want to inspect
        print_structure: (optional, default = True)
        indent: (optional, default = 0) indent of the first output

    Returns:
        Dictionary that has the same structure as the input_dict, but as values it has the
            type(input_dict[key]) (except if the value is a dict)
    """

    structure_dict = collections.OrderedDict()  # TODO: check if this still works

    #
    for key, value in input_dict.items():
        structure_dict[key] = type(value)

        if print_structure:
            print("\t" * indent, key, type(value))

        if isinstance(value, dict):
            structure_dict[key] = inspect_dict(
                value, indent=indent + 1, print_structure=print_structure
            )

    return structure_dict


def count_keys_recursive(input_dict):
    """
    Function to count the total number of keys in a dictionary
    """

    local_count = 0
    for key in input_dict.keys():
        local_count += 1
        if isinstance(input_dict[key], (dict, collections.OrderedDict)):
            local_count += count_keys_recursive(input_dict[key])
    return local_count


def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
    """
    Function to merge two dictionaries in a custom way.

    Behaviour:

    When dict keys are only present in one of either:
        - we just add the content to the new dict

    When dict keys are present in both, we decide based on the value types how to combine them:
        - dictionaries will be merged by calling recursively calling this function again
        - numbers will be added
        - (opt) lists will be appended
        - In the case that the instances do not match: for now I will raise an error

    Args:
        dict_1: first dictionary
        dict_2: second dictionary

    Returns:
        Merged dictionary

    """

    # Set up new dict
    new_dict = collections.OrderedDict()  # TODO: check if this still necessary

    #
    keys_1 = dict_1.keys()
    keys_2 = dict_2.keys()

    # Find overlapping keys of both dicts
    overlapping_keys = set(keys_1).intersection(set(keys_2))

    # Find the keys that are unique
    unique_to_dict_1 = set(keys_1).difference(set(keys_2))
    unique_to_dict_2 = set(keys_2).difference(set(keys_1))

    # Add the unique keys to the new dict
    for key in unique_to_dict_1:
        # If these items are ints or floats, then just put them in
        if isinstance(dict_1[key], (float, int)):
            new_dict[key] = dict_1[key]
        # Else, to be safe we should deepcopy them
        else:
            copy_dict = copy.deepcopy(dict_1[key])
            new_dict[key] = copy_dict

    for key in unique_to_dict_2:
        # If these items are ints or floats, then just put them in
        if isinstance(dict_2[key], (float, int)):
            new_dict[key] = dict_2[key]
        # Else, to be safe we should deepcopy them
        else:
            copy_dict = copy.deepcopy(dict_2[key])
            new_dict[key] = copy_dict

    # Go over the common keys:
    for key in overlapping_keys:

        # If they keys are not the same, it depends on their type whether we still deal with them at all, or just raise an error
        if not type(dict_1[key]) is type(dict_2[key]):
            # Exceptions: numbers can be added
            if isinstance(dict_1[key], (int, float, np.float64)) and isinstance(
                dict_2[key], (int, float, np.float64)
            ):
                new_dict[key] = dict_1[key] + dict_2[key]

            # Exceptions: versions of dicts can be merged
            elif isinstance(
                dict_1[key], (dict, collections.OrderedDict, type(AutoVivificationDict))
            ) and isinstance(
                dict_2[key], (dict, collections.OrderedDict, type(AutoVivificationDict))
            ):
                new_dict[key] = merge_dicts(dict_1[key], dict_2[key])

            # If the above cases have not dealt with it, then we should raise an error
            else:
                print(
                    "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
                        key,
                        dict_1[key],
                        type(dict_1[key]),
                        key,
                        dict_2[key],
                        type(dict_2[key]),
                    )
                )
                raise ValueError

        # Here we check for the cases that we want to explicitly catch. Ints will be added,
        # floats will be added, lists will be appended (though that might change) and dicts will be
        # dealt with by calling this function again.
        else:
            # ints
            # Booleans (has to be the type Bool, not just a 0 or 1)
            if isinstance(dict_1[key], bool) and isinstance(dict_2[key], bool):
                new_dict[key] = dict_1[key] or dict_2[key]

            elif isinstance(dict_1[key], int) and isinstance(dict_2[key], int):
                new_dict[key] = dict_1[key] + dict_2[key]

            # floats
            elif isinstance(dict_1[key], float) and isinstance(dict_2[key], float):
                new_dict[key] = dict_1[key] + dict_2[key]

            # lists
            elif isinstance(dict_1[key], list) and isinstance(dict_2[key], list):
                new_dict[key] = dict_1[key] + dict_2[key]

            # Astropy quantities (using a dummy type representing the numpy array)
            elif isinstance(dict_1[key], type(np.array([1]) * u.m)) and isinstance(
                dict_2[key], type(np.array([1]) * u.m)
            ):
                new_dict[key] = dict_1[key] + dict_2[key]

            # dicts
            elif isinstance(dict_1[key], dict) and isinstance(dict_2[key], dict):
                new_dict[key] = merge_dicts(dict_1[key], dict_2[key])

            else:
                print(
                    "Object types {}: {} ({}), {} ({}) not supported.".format(
                        key,
                        dict_1[key],
                        type(dict_1[key]),
                        dict_2[key],
                        type(dict_2[key]),
                    )
                )
                raise ValueError

    #
    return new_dict


def update_dicts(dict_1: dict, dict_2: dict) -> dict:
    """
    Function to update dict_1 with values of dict_2 in a recursive way.

    Behaviour:

    When dict keys are only present in one of either:
        - we just add the content to the new dict

    When dict keys are present in both, we decide based on the value types how to combine them:
        - value of dict2 will be taken

    Args:
        dict_1: first dictionary
        dict_2: second dictionary

    Returns:
        New dictionary with Updated values

    """

    # Set up new dict of the same type as dict_1
    #
    # Note: setting directly to OrderedDict fails in some cases
    # so instead we take a (shallow) copy of dict_1 which will
    # have the same type as dict_1, then clear it. (There must
    # be a better way to do this...)
    new_dict = dict_1.copy()  # OrderedDict()  # TODO: check if this still works
    new_dict.clear()

    keys_1 = dict_1.keys()
    keys_2 = dict_2.keys()

    # Find overlapping keys of both dicts
    overlapping_keys = set(keys_1).intersection(set(keys_2))

    # Find the keys that are unique
    unique_to_dict_1 = set(keys_1).difference(set(keys_2))
    unique_to_dict_2 = set(keys_2).difference(set(keys_1))

    # Add the unique keys to the new dict
    for key in unique_to_dict_1:
        # If these items are ints or floats, then just put them in
        if isinstance(dict_1[key], (float, int)):
            new_dict[key] = dict_1[key]
        # Else, to be safe we should deepcopy them
        else:
            copy_dict = copy.deepcopy(dict_1[key])
            new_dict[key] = copy_dict

    for key in unique_to_dict_2:
        # If these items are ints or floats, then just put them in
        if isinstance(dict_2[key], (float, int)):
            new_dict[key] = dict_2[key]
        # Else, to be safe we should deepcopy them
        else:
            copy_dict = copy.deepcopy(dict_2[key])
            new_dict[key] = copy_dict

    # Go over the common keys:
    for key in overlapping_keys:

        # See whether the types are actually the same
        if not type(dict_1[key]) is type(dict_2[key]):
            # Exceptions:
            if (type(dict_1[key]) in [int, float]) and (
                type(dict_2[key]) in [int, float]
            ):
                new_dict[key] = dict_2[key]

            else:
                print(
                    "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
                        key,
                        dict_1[key],
                        type(dict_1[key]),
                        key,
                        dict_2[key],
                        type(dict_2[key]),
                    )
                )
                raise ValueError

        # Here we check for the cases that we want to explicitly catch. Ints will be added,
        # floats will be added, lists will be appended (though that might change) and dicts will be
        # dealt with by calling this function again.
        else:
            # dicts
            if isinstance(dict_1[key], dict) and isinstance(dict_2[key], dict):
                new_dict[key] = update_dicts(dict_1[key], dict_2[key])
            else:
                new_dict[key] = dict_2[key]

    #
    return new_dict


def multiply_values_dict(input_dict, factor):
    """
    Function that goes over dictionary recursively and multiplies the value if possible by a factor

    If the key equals "general_info", the multiplication gets skipped
    """

    for key in input_dict:
        if not key == "general_info":
            if isinstance(input_dict[key], (dict, collections.OrderedDict)):
                input_dict[key] = multiply_values_dict(input_dict[key], factor)
            else:
                if isinstance(input_dict[key], (int, float)):
                    input_dict[key] = input_dict[key] * factor

    return input_dict


def custom_sort_dict(input_dict):
    """
    Returns a dictionary that is ordered, but can handle numbers better than normal OrderedDict

    When the keys of the current dictionary are of mixed type, we first find all the unique types.
    Sort that list of type names. Then find the values that fit that type.
    Sort those and append them to the sorted keys list.
    This is done until all the keys are sorted.

    All objects other than dictionary types are directly return as they are
    """

    # If the new input is a dictionary, then try to sort it
    if isinstance(input_dict, (dict, collections.OrderedDict)):
        new_dict = collections.OrderedDict()

        keys = input_dict.keys()

        # Check if types are the same
        all_types_keys = []
        for key in keys:
            if not type(key) in all_types_keys:
                all_types_keys.append(type(key))

        # If there are multiple types, then we loop over them and do a piece wise sort
        if len(all_types_keys) > 1:
            msg = "Different types in the same dictionary key set"

            # Create a string repr of the type name to sort them afterwards
            str_types = {repr(el): el for el in all_types_keys}

            # Set up sorted keys list
            sorted_keys = []

            for key_str_type in sorted(str_types.keys()):
                cur_type = str_types[key_str_type]

                cur_list = [key for key in keys if isinstance(key, cur_type)]
                cur_sorted_list = sorted(cur_list)

                sorted_keys = sorted_keys + cur_sorted_list
        else:
            sorted_keys = sorted(keys)

        for key in sorted_keys:
            new_dict[key] = custom_sort_dict(copy.deepcopy(input_dict[key]))

        return new_dict
    return input_dict


def filter_dict(arg_dict: dict, filter_list: list) -> dict:
    """
    Function to filter out keys that are contains in filter_list

    Args:
        arg_dict: dictionary containing the argument + default key pairs of binary_c
        filter_list: lists of keys to be filtered out
    Returns:
        filtered dictionary
    """

    new_dict = arg_dict.copy()

    for key in filter_list:
        if key in new_dict:
            del new_dict[key]

    return new_dict


def filter_dict_through_values(arg_dict: dict, filter_list: list) -> dict:
    """
    Function to filter out keys that contain values included in filter_list

    Args:
        arg_dict: dictionary containing the argument + default key pairs of binary_c
        filter_list: lists of values to be filtered out
    Returns:
        filtered dictionary
    """

    new_dict = {}

    for key in arg_dict:
        if not arg_dict[key] in filter_list:
            new_dict[key] = arg_dict[key]

    return new_dict



import unittest


import nbformat
import os

from nbconvert.preprocessors import ExecutePreprocessor

from binarycpython.utils.functions import temp_dir

TMP_DIR = temp_dir("testing", "test_notebooks")
NOTEBOOKS_DIR = os.path.abspath(
    os.path.join(os.path.abspath(__file__), "../../../examples")
)


def run_notebook(notebook_path):
    """
    Function to run notebooks and get the errors
    """

    # https://www.blog.pythonlibrary.org/2018/10/16/testing-jupyter-notebooks/
    nb_name, _ = os.path.splitext(os.path.basename(notebook_path))
    dirname = os.path.dirname(notebook_path)

    with open(notebook_path) as f:
        nb = nbformat.read(f, as_version=4)

    proc = ExecutePreprocessor(timeout=600, kernel_name="python3")
    proc.allow_errors = True

    proc.preprocess(nb, {"metadata": {"path": "/"}})

    output_path = os.path.join(TMP_DIR, "{}_all_output.ipynb".format(nb_name))

    with open(output_path, mode="wt") as f:
        nbformat.write(nb, f)
    errors = []
    for cell in nb.cells:
        if "outputs" in cell:
            for output in cell["outputs"]:
                if output.output_type == "error":
                    errors.append(output)

    return nb, errors


class TestNotebook(unittest.TestCase):
    """
    Class that contains the notebook test calls
    """

    def test_notebook_api_functionality(self):
        notebook_name = "notebook_api_functionality.ipynb"
        full_notebook_path = os.path.join(NOTEBOOKS_DIR, notebook_name)
        nb, errors = run_notebook(full_notebook_path)
        msg = "\nNotebook: {}\n\n".format(notebook_name) + "\n".join(
            [
                "{}: {}\n{}".format(
                    el["ename"], el["evalue"], "\n".join(el["traceback"])
                )
                for el in errors
            ]
        )
        self.assertEqual(errors, [], msg=msg)

    def test_notebook_population(self):
        notebook_name = "notebook_population.ipynb"
        full_notebook_path = os.path.join(NOTEBOOKS_DIR, notebook_name)
        nb, errors = run_notebook(full_notebook_path)
        msg = "\nNotebook: {}\n\n".format(notebook_name) + "\n".join(
            [
                "{}: {}\n{}".format(
                    el["ename"], el["evalue"], "\n".join(el["traceback"])
                )
                for el in errors
            ]
        )
        self.assertEqual(errors, [], msg=msg)

    def test_notebook_individual_systems(self):
        notebook_name = "notebook_individual_systems.ipynb"
        full_notebook_path = os.path.join(NOTEBOOKS_DIR, notebook_name)
        nb, errors = run_notebook(full_notebook_path)
        msg = "\nNotebook: {}\n\n".format(notebook_name) + "\n".join(
            [
                "{}: {}\n{}".format(
                    el["ename"], el["evalue"], "\n".join(el["traceback"])
                )
                for el in errors
            ]
        )
        self.assertEqual(errors, [], msg=msg)

    def test_notebook_custom_logging(self):
        notebook_name = "notebook_custom_logging.ipynb"
        full_notebook_path = os.path.join(NOTEBOOKS_DIR, notebook_name)
        nb, errors = run_notebook(full_notebook_path)
        msg = "\nNotebook: {}\n\n".format(notebook_name) + "\n".join(
            [
                "{}: {}\n{}".format(
                    el["ename"], el["evalue"], "\n".join(el["traceback"])
                )
                for el in errors
            ]
        )
        self.assertEqual(errors, [], msg=msg)

    def test_notebook_extra_features(self):
        notebook_name = "notebook_extra_features.ipynb"
        full_notebook_path = os.path.join(NOTEBOOKS_DIR, notebook_name)
        nb, errors = run_notebook(full_notebook_path)
        msg = "\nNotebook: {}\n\n".format(notebook_name) + "\n".join(
            [
                "{}: {}\n{}".format(
                    el["ename"], el["evalue"], "\n".join(el["traceback"])
                )
                for el in errors
            ]
        )
        self.assertEqual(errors, [], msg=msg)

    def test_notebook_luminosity_function_single(self):
        notebook_name = "notebook_luminosity_function_single.ipynb"
        full_notebook_path = os.path.join(NOTEBOOKS_DIR, notebook_name)
        nb, errors = run_notebook(full_notebook_path)
        msg = "\nNotebook: {}\n\n".format(notebook_name) + "\n".join(
            [
                "{}: {}\n{}".format(
                    el["ename"], el["evalue"], "\n".join(el["traceback"])
                )
                for el in errors
            ]
        )
        self.assertEqual(errors, [], msg=msg)


if __name__ == "__main__":
    unittest.main()



"""
Module containing the predefined distribution functions

The user can use any of these distribution functions to
generate probability distributions for sampling populations

There are distributions for the following parameters:
    - mass
    - period
    - mass ratio
    - binary fraction

Tasks:
    - TODO: make some things globally present? rob does this in his module..i guess it saves
        calculations but not sure if I'm gonna do that now
    - TODO: add eccentricity distribution: thermal
    - TODO: Add SFH distributions depending on redshift
    - TODO: Add metallicity distributions depending on redshift
    - TODO: Add initial rotational velocity distributions
    - TODO: make an n-part power law that's general enough to fix the three part and the 4 part
"""

import functools
import math
import json

import traceback
import sys

from typing import Union

import numpy as np

from binarycpython.utils.useful_funcs import calc_period_from_sep, calc_sep_from_period
from binarycpython.utils.functions import verbose_print
from binarycpython.utils.grid_options_defaults import (
    _MOE2017_VERBOSITY_LEVEL,
    _MOE2017_VERBOSITY_INTERPOLATOR_LEVEL,
)

###
# File containing probability distributions
# Mostly copied from the Perl modules
LOG_LN_CONVERTER = 1.0 / math.log(10.0)
distribution_constants = {}  # To store the constants in


def prepare_dict(global_dict: dict, list_of_sub_keys: list) -> None:
    """
    Function that makes sure that the global dict is prepared to have a value set there.
    This dictionary will store values and factors for the distribution functions,
    so that they don't have to be calculated each time.

    Args:
        global_dict: globally accessible dictionary where factors are stored in
        list_of_sub_keys: List of keys that must become be(come) present in the global_dict
    """

    internal_dict_value = global_dict

    # This loop almost mimics a recursive loop into the dictionary.
    # It checks whether the first key of the list is present, if not; set it with an empty dict.
    # Then it overrides itself to be that (new) item, and goes on to do that again, until the list
    # exhausted
    for k in list_of_sub_keys:
        # If the sub key doesnt exist then make an empty dict
        if not internal_dict_value.get(k, None):
            internal_dict_value[k] = {}
        internal_dict_value = internal_dict_value[k]


def set_opts(opts: dict, newopts: dict) -> dict:
    """
    Function to take a default dict and override it with newer values.

    # TODO: consider changing this to just a dict.update

    Args:
        opts: dictionary with default values
        newopts: dictionary with new values

    Returns:
        returns an updated dictionary
    """

    if newopts:
        for opt in newopts.keys():
            if opt in opts.keys():
                opts[opt] = newopts[opt]

    return opts


def flat() -> float:
    """
    Dummy distribution function that returns 1

    Returns:
        a flat uniform distribution: 1
    """

    return 1.0


def number(value: Union[int, float]) -> Union[int, float]:
    """
    Dummy distribution function that returns the input

    Args:
        value: the value that will be returned by this function.

    Returns:
        the value that was provided
    """

    return value


def const(
    min_bound: Union[int, float], max_bound: Union[int, float], val: float = None
) -> Union[int, float]:
    """
    a constant distribution function between min=min_bound and max=max_bound.

    Args:
        min_bound: lower bound of the range
        max_bound: upper bound of the range

    Returns:
            returns the value of 1/(max_bound-min_bound). If val is provided, it will check whether min_bound < val <= max_bound. if not: returns 0
    """

    if val:
        if not min_bound < val <= max_bound:
            print("out of bounds")
            prob = 0
            return prob
    prob = 1.0 / (max_bound - min_bound)
    return prob


@functools.lru_cache(maxsize=16)
def powerlaw_constant(
    min_val: Union[int, float], max_val: Union[int, float], k: Union[int, float]
) -> Union[int, float]:
    """
    Function that returns the constant to normalise a power law

    TODO: what if k is -1?

    Args:
        min_val: lower bound of the range
        max_val: upper bound of the range
        k: power law slope

    Returns:
        constant to normalise the given power law between the min_val and max_val range
    """

    k1 = k + 1.0
    # print(
    #     "Powerlaw consts from {} to {}, k={} where k1={}".format(
    #         min_val, max_val, k, k1
    #     )
    # )

    powerlaw_const = k1 / (max_val ** k1 - min_val ** k1)
    return powerlaw_const


def powerlaw(
    min_val: Union[int, float],
    max_val: Union[int, float],
    k: Union[int, float],
    x: Union[int, float],
) -> Union[int, float]:
    """
    Single power law with index k at x from min to max

    Args:
        min_val: lower bound of the power law
        max_val: upper bound of the power law
        k: slope of the power law
        x: position at which we want to evaluate

    Returns:
        `probability` at the given position(x)
    """

    # Handle faulty value
    if k == -1:
        msg = "wrong value for k"
        raise ValueError(msg)

    if (x < min_val) or (x > max_val):
        print("input value is out of bounds!")
        return 0

    powerlaw_const = powerlaw_constant(min_val, max_val, k)

    # power law
    prob = powerlaw_const * (x ** k)
    # print(
    #     "Power law from {} to {}: const = {}, y = {}".format(
    #         min_val, max_val, const, y
    #     )
    # )
    return prob


@functools.lru_cache(maxsize=16)
def calculate_constants_three_part_powerlaw(
    m0: Union[int, float],
    m1: Union[int, float],
    m2: Union[int, float],
    m_max: Union[int, float],
    p1: Union[int, float],
    p2: Union[int, float],
    p3: Union[int, float],
) -> Union[int, float]:
    """
    Function to calculate the constants for a three-part power law

    TODO: use the power law_constant function to calculate all these values

    Args:
        m0: lower bound mass
        m1: second boundary, between the first slope and the second slope
        m2: third boundary, between the second slope and the third slope
        m_max: upper bound mass
        p1: first slope
        p2: second slope
        p3: third slope

    Returns:
        array of normalisation constants
    """

    # print("Initialising constants for the three-part powerlaw: m0={} m1={} m2={}\
    # m_max={} p1={} p2={} p3={}\n".format(m0, m1, m2, m_max, p1, p2, p3))

    array_constants_three_part_powerlaw = [0, 0, 0]

    array_constants_three_part_powerlaw[1] = (
        ((m1 ** p2) * (m1 ** (-p1)))
        * (1.0 / (1.0 + p1))
        * (m1 ** (1.0 + p1) - m0 ** (1.0 + p1))
    )
    array_constants_three_part_powerlaw[1] += (
        (m2 ** (1.0 + p2) - m1 ** (1.0 + p2))
    ) * (1.0 / (1.0 + p2))
    array_constants_three_part_powerlaw[1] += (
        ((m2 ** p2) * (m2 ** (-p3)))
        * (1.0 / (1.0 + p3))
        * (m_max ** (1.0 + p3) - m2 ** (1.0 + p3))
    )
    array_constants_three_part_powerlaw[1] = 1.0 / (
        array_constants_three_part_powerlaw[1] + 1e-50
    )

    array_constants_three_part_powerlaw[0] = array_constants_three_part_powerlaw[1] * (
        (m1 ** p2) * (m1 ** (-p1))
    )
    array_constants_three_part_powerlaw[2] = array_constants_three_part_powerlaw[1] * (
        (m2 ** p2) * (m2 ** (-p3))
    )

    return array_constants_three_part_powerlaw
    # $$array[1]=(($m1**$p2)*($m1**(-$p1)))*
    # (1.0/(1.0+$p1))*
    # ($m1**(1.0+$p1)-$m0**(1.0+$p1))+
    # (($m2**(1.0+$p2)-$m1**(1.0+$p2)))*
    # (1.0/(1.0+$p2))+
    # (($m2**$p2)*($m2**(-$p3)))*
    # (1.0/(1.0+$p3))*
    # ($mmax**(1.0+$p3)-$m2**(1.0+$p3));
    # $$array[1]=1.0/($$array[1]+1e-50);
    # $$array[0]=$$array[1]*$m1**$p2*$m1**(-$p1);
    # $$array[2]=$$array[1]*$m2**$p2*$m2**(-$p3);
    # #print "ARRAY SET @_ => @$array\n";
    # $threepart_powerlaw_consts{"@_"}=[@$array];


def three_part_powerlaw(
    m: Union[int, float],
    m0: Union[int, float],
    m1: Union[int, float],
    m2: Union[int, float],
    m_max: Union[int, float],
    p1: Union[int, float],
    p2: Union[int, float],
    p3: Union[int, float],
) -> Union[int, float]:
    """
    Generalised three-part power law, usually used for mass distributions

    Args:
        m: mass at which we want to evaluate the distribution.
        m0: lower bound mass
        m1: second boundary, between the first slope and the second slope
        m2: third boundary, between the second slope and the third slope
        m_max: upper bound mass
        p1: first slope
        p2: second slope
        p3: third slope

    Returns:
        'probability' at given mass m
    """

    # TODO: add check on whether the values exist

    three_part_powerlaw_constants = calculate_constants_three_part_powerlaw(
        m0, m1, m2, m_max, p1, p2, p3
    )

    #
    if m < m0:
        prob = 0  # Below lower bound TODO: make this clear.
    elif m0 < m <= m1:
        prob = three_part_powerlaw_constants[0] * (m ** p1)  # Between M0 and M1
    elif m1 < m <= m2:
        prob = three_part_powerlaw_constants[1] * (m ** p2)  # Between M1 and M2
    elif m2 < m <= m_max:
        prob = three_part_powerlaw_constants[2] * (m ** p3)  # Between M2 and M_MAX
    else:
        prob = 0  # Above M_MAX

    return prob


@functools.lru_cache(maxsize=16)
def gaussian_normalizing_const(
    mean: Union[int, float],
    sigma: Union[int, float],
    gmin: Union[int, float],
    gmax: Union[int, float],
) -> Union[int, float]:
    """
    Function to calculate the normalisation constant for the Gaussian

    Args:
        mean: mean of the Gaussian
        sigma: standard deviation of the Gaussian
        gmin: lower bound of the range to calculate the probabilities in
        gmax: upper bound of the range to calculate the probabilities in

    Returns:
        normalisation constant for the Gaussian distribution(mean, sigma) between gmin and gmax
    """

    # First time; calculate multiplier for given mean and sigma
    ptot = 0
    resolution = 1000
    d = (gmax - gmin) / resolution

    for i in range(resolution):
        y = gmin + i * d
        ptot += d * gaussian_func(y, mean, sigma)

    # TODO: Set value in global
    return ptot


def gaussian_func(
    x: Union[int, float], mean: Union[int, float], sigma: Union[int, float]
) -> Union[int, float]:
    """
    Function to evaluate a Gaussian at a given point, but this time without any boundaries.

    Args:
        x: location at which to evaluate the distribution
        mean: mean of the Gaussian
        sigma: standard deviation of the Gaussian

    Returns:
        value of the Gaussian at x
    """
    gaussian_prefactor = 1.0 / math.sqrt(2.0 * math.pi)

    r = 1.0 / (sigma)
    y = (x - mean) * r
    return gaussian_prefactor * r * math.exp(-0.5 * y ** 2)


def gaussian(
    x: Union[int, float],
    mean: Union[int, float],
    sigma: Union[int, float],
    gmin: Union[int, float],
    gmax: Union[int, float],
) -> Union[int, float]:
    """
    Gaussian distribution function. used for e.g. Duquennoy + Mayor 1991

    Args:
        x: location at which to evaluate the distribution
        mean: mean of the Gaussian
        sigma: standard deviation of the Gaussian
        gmin: lower bound of the range to calculate the probabilities in
        gmax: upper bound of the range to calculate the probabilities in

    Returns:
        'probability' of the Gaussian distribution between the boundaries, evaluated at x
    """

    # # location (X value), mean and sigma, min and max range
    # my ($x,$mean,$sigma,$gmin,$gmax) = @_;

    if (x < gmin) or (x > gmax):
        prob = 0
    else:
        # normalise over given range
        # TODO: add loading into global var
        normalisation = gaussian_normalizing_const(mean, sigma, gmin, gmax)
        prob = gaussian_func(x, mean, sigma) / normalisation

    return prob


#####
# Mass distributions
#####


def Kroupa2001(m: Union[int, float], newopts: dict = None) -> Union[int, float]:
    """
    Probability distribution function for Kroupa 2001 IMF, where the default values to the
    three_part_powerlaw are: default = {"m0": 0.1, "m1": 0.5, "m2": 1, "mmax": 100, "p1": -1.3, "p2": -2.3,"p3": -2.3}

    Args:
        m: mass to evaluate the distribution at
        newopts: optional dict to override the default values.

    Returns:
        'probability' of distribution function evaluated at m
    """

    # Default parameters and override them
    default = {
        "m0": 0.1,
        "m1": 0.5,
        "m2": 1,
        "mmax": 100,
        "p1": -1.3,
        "p2": -2.3,
        "p3": -2.3,
    }

    value_dict = default.copy()

    if newopts:
        value_dict.update(newopts)

    return three_part_powerlaw(
        m,
        value_dict["m0"],
        value_dict["m1"],
        value_dict["m2"],
        value_dict["mmax"],
        value_dict["p1"],
        value_dict["p2"],
        value_dict["p3"],
    )


def ktg93(m: Union[int, float], newopts: dict = None) -> Union[int, float]:
    """
    Probability distribution function for KTG93 IMF, where the default values to the three_part_powerlaw are: default = {"m0": 0.1, "m1": 0.5, "m2": 1, "mmax": 80, "p1": -1.3, "p2": -2.2,"p3": -2.7}

    Args:
        m: mass to evaluate the distribution at
        newopts: optional dict to override the default values.

    Returns:
        'probability' of distribution function evaluated at m
    """
    # TODO: ask rob what this means

    # if($m eq 'uncertainties')
    # {
    # # return (pointer to) the uncertainties hash
    # return {
    #     m0=>{default=>0.1,
    #      fixed=>1},
    #     m1=>{default=>0.5,
    #      fixed=>1},
    #     m2=>{default=>1.0,
    #      fixed=>1},
    #     mmax=>{default=>80.0,
    #        fixed=>1},
    #     p1=>{default=>-1.3,
    #      low=>-1.3,
    #      high=>-1.3},
    #     p2=>{default=>-2.2,
    #      low=>-2.2,
    #      high=>-2.2},
    #     p3=>{default=>-2.7,
    #      low=>-2.7,
    #      high=>-2.7}
    # };
    # }

    # set options
    # opts = set_opts({'m0':0.1, 'm1':0.5, 'm2':1.0, 'mmax':80, 'p1':-1.3, 'p2':-2.2, 'p3':-2.7},
    # newopts)

    defaults = {
        "m0": 0.1,
        "m1": 0.5,
        "m2": 1.0,
        "mmax": 80,
        "p1": -1.3,
        "p2": -2.2,
        "p3": -2.7,
    }
    value_dict = defaults.copy()

    if newopts:
        value_dict.update(newopts)

    return three_part_powerlaw(
        m,
        value_dict["m0"],
        value_dict["m1"],
        value_dict["m2"],
        value_dict["mmax"],
        value_dict["p1"],
        value_dict["p2"],
        value_dict["p3"],
    )


# sub ktg93_lnspace
# {
#     # wrapper for KTG93 on a ln(m) grid
#     my $m=$_[0];
#     return ktg93(@_) * $m;
# }


def imf_tinsley1980(m: Union[int, float]) -> Union[int, float]:
    """
    Probability distribution function for Tinsley 1980 IMF (defined up until 80Msol): three_part_powerlaw(m, 0.1, 2.0, 10.0, 80.0, -2.0, -2.3, -3.3)

    Args:
        m: mass to evaluate the distribution at

    Returns:
        'probability' of distribution function evaluated at m
    """

    return three_part_powerlaw(m, 0.1, 2.0, 10.0, 80.0, -2.0, -2.3, -3.3)


def imf_scalo1986(m: Union[int, float]) -> Union[int, float]:
    """
    Probability distribution function for Scalo 1986 IMF (defined up until 80Msol): three_part_powerlaw(m, 0.1, 1.0, 2.0, 80.0, -2.35, -2.35, -2.70)

    Args:
        m: mass to evaluate the distribution at

    Returns:
        'probability' of distribution function evaluated at m
    """
    return three_part_powerlaw(m, 0.1, 1.0, 2.0, 80.0, -2.35, -2.35, -2.70)


def imf_scalo1998(m: Union[int, float]) -> Union[int, float]:
    """
    From Scalo 1998

    Probability distribution function for Scalo 1998 IMF (defined up until 80Msol): three_part_powerlaw(m, 0.1, 1.0, 10.0, 80.0, -1.2, -2.7, -2.3)

    Args:
        m: mass to evaluate the distribution at

    Returns:
        'probability' of distribution function evaluated at m
    """

    return three_part_powerlaw(m, 0.1, 1.0, 10.0, 80.0, -1.2, -2.7, -2.3)


def imf_chabrier2003(m: Union[int, float]) -> Union[int, float]:
    """
    Probability distribution function for IMF of Chabrier 2003 PASP 115:763-795

    Args:
        m: mass to evaluate the distribution at

    Returns:
        'probability' of distribution function evaluated at m
    """

    chabrier_logmc = math.log10(0.079)
    chabrier_sigma2 = 0.69 * 0.69
    chabrier_a1 = 0.158
    chabrier_a2 = 4.43e-2
    chabrier_x = -1.3
    if m <= 0:
        msg = "below bounds"
        raise ValueError(msg)
    if 0 < m < 1.0:
        A = 0.158
        dm = math.log10(m) - chabrier_logmc
        prob = chabrier_a1 * math.exp(-(dm ** 2) / (2.0 * chabrier_sigma2))
    else:
        prob = chabrier_a2 * (m ** chabrier_x)
    prob = prob / (0.1202462 * m * math.log(10))
    return prob


########################################################################
# Binary fractions
########################################################################


def Arenou2010_binary_fraction(m: Union[int, float]) -> Union[int, float]:
    """
    Arenou 2010 function for the binary fraction as f(M1)

    GAIA-C2-SP-OPM-FA-054
    www.rssd.esa.int/doc_fetch.php?id=2969346

    Args:
        m: mass to evaluate the distribution at

    Returns:
        binary fraction at m
    """

    return 0.8388 * math.tanh(0.688 * m + 0.079)


# print(Arenou2010_binary_fraction(0.4))


def raghavan2010_binary_fraction(m: Union[int, float]) -> Union[int, float]:
    """
    Fit to the Raghavan 2010 binary fraction as a function of
    spectral type (Fig 12). Valid for local stars (Z=Zsolar).

    The spectral type is converted  mass by use of the ZAMS
    effective temperatures from binary_c/BSE (at Z=0.02)
    and the new "long_spectral_type" function of binary_c
    (based on Jaschek+Jaschek's Teff-spectral type table).

    Rob then fitted the result

    Args:
        m: mass to evaluate the distribution at

    Returns:
        binary fraction at m
    """

    return min(
        1.0,
        max(
            (m ** 0.1) * (5.12310e-01) + (-1.02070e-01),
            (1.10450e00) * (m ** (4.93670e-01)) + (-6.95630e-01),
        ),
    )


# print(raghavan2010_binary_fraction(2))

########################################################################
# Period distributions
########################################################################


def duquennoy1991(logper: Union[int, float]) -> Union[int, float]:
    """
    Period distribution from Duquennoy + Mayor 1991. Evaluated the function gaussian(logper, 4.8, 2.3, -2, 12)

    Args:
        logper: logarithm of period to evaluate the distribution at

    Returns:
        'probability' at gaussian(logper, 4.8, 2.3, -2, 12)
    """
    return gaussian(logper, 4.8, 2.3, -2, 12)


def sana12(
    M1: Union[int, float],
    M2: Union[int, float],
    a: Union[int, float],
    P: Union[int, float],
    amin: Union[int, float],
    amax: Union[int, float],
    x0: Union[int, float],
    x1: Union[int, float],
    p: Union[int, float],
) -> Union[int, float]:
    """
    distribution of initial orbital periods as found by Sana et al. (2012)
    which is a flat distribution in ln(a) and ln(P) respectively for stars
    * less massive than 15Msun (no O-stars)
    * mass ratio q=M2/M1<0.1
    * log(P)<0.15=x0 and log(P)>3.5=x1
    and is be given by dp/dlogP ~ (logP)^p for all other binary configurations (default p=-0.55)

    arguments are M1, M2, a, Period P, amin, amax, x0=log P0, x1=log P1, p

    example args: 10, 5, sep(M1, M2, P), sep, ?, -2, 12, -0.55

    # TODO: Fix this function!

    Args:
        M1: Mass of primary
        M2: Mass of secondary
        a: separation of binary
        P: period of binary
        amin: minimum separation of the distribution (lower bound of the range)
        amax: maximum separation of the distribution (upper bound of the range)
        x0: log of minimum period of the distribution (lower bound of the range)
        x1: log of maximum period of the distribution (upper bound of the range)
        p: slope of the distribution

    Returns:
        'probability' of orbital period P given the other parameters
    """

    res = 0
    if (M1 < 15.0) or (M2 / M1 < 0.1):
        res = 1.0 / (math.log(amax) - math.log(amin))
    else:
        p1 = 1.0 + p

        # For more details see the LyX document of binary_c for this distribution
        # where the variables and normalisations are given
        # we use the notation x=log(P), xmin=log(Pmin), x0=log(P0), ... to determine the
        x = LOG_LN_CONVERTER * math.log(P)
        xmin = LOG_LN_CONVERTER * math.log(calc_period_from_sep(M1, M2, amin))
        xmax = LOG_LN_CONVERTER * math.log(calc_period_from_sep(M1, M2, amax))

        # print("M1 M2 amin amax P x xmin xmax")
        # print(M1, M2, amin, amax, P, x, xmin, xmax)
        # my $x0 = 0.15;
        # my $x1 = 3.5;

        A1 = 1.0 / (
            x0 ** p * (x0 - xmin) + (x1 ** p1 - x0 ** p1) / p1 + x1 ** p * (xmax - x1)
        )
        A0 = A1 * x0 ** p
        A2 = A1 * x1 ** p

        if x < x0:
            res = 3.0 / 2.0 * LOG_LN_CONVERTER * A0
        elif x > x1:
            res = 3.0 / 2.0 * LOG_LN_CONVERTER * A2
        else:
            res = 3.0 / 2.0 * LOG_LN_CONVERTER * A1 * x ** p

    return res


# print(sana12(10, 2, 10, 100, 1, 1000, math.log(10), math.log(1000), 6))


def interpolate_in_mass_izzard2012(
    M: Union[int, float], high: Union[int, float], low: Union[int, float]
) -> Union[int, float]:
    """
    Function to interpolate in mass

    TODO: fix this function.
    TODO: describe the args
    high: at M=16.3
    low: at 1.15

    Args:
        M: mass
        high:
        low:

    Returns:

    """

    log_interpolation = False

    if log_interpolation:
        return (high - low) / (math.log10(16.3) - math.log10(1.15)) * (
            math.log10(M) - math.log10(1.15)
        ) + low
    else:
        return (high - low) / (16.3 - 1.15) * (M - 1.15) + low


def Izzard2012_period_distribution(
    P: Union[int, float], M1: Union[int, float], log10Pmin: Union[int, float] = -1.0
) -> Union[int, float]:
    """
    period distribution which interpolates between
    Duquennoy and Mayor 1991 at low mass (G/K spectral type <~1.15Msun)
    and Sana et al 2012 at high mass (O spectral type >~16.3Msun)

    This gives dN/dlogP, i.e. DM/Raghavan's Gaussian in log10P at low mass
    and Sana's power law (as a function of logP) at high mass

    TODO: fix this function

    Args:
        P: period
        M1: Primary star mass
        log10Pmin: minimum period in base log10 (optional)

    Returns:
        'probability' of interpolated distribution function at P and M1

    """

    # Check if there is input and force it to be at least 1
    log10Pmin = max(-1.0, log10Pmin)

    # save mass input and limit mass used (M1 from now on) to fitted range
    Mwas = M1
    M1 = max(1.15, min(16.3, M1))
    # print("Izzard2012 called for M={} (truncated to {}), P={}\n".format(Mwas, M1, P))

    # Calculate the normalisations
    # need to normalise the distribution for this mass
    # (and perhaps secondary mass)
    prepare_dict(distribution_constants, ["Izzard2012", M1])
    if not distribution_constants["Izzard2012"][M1].get(log10Pmin):
        distribution_constants["Izzard2012"][M1][
            log10Pmin
        ] = 1  # To prevent this loop from going recursive
        N = 200.0  # Resolution for normalisation. I hope 1000 is enough
        dlP = (10.0 - log10Pmin) / N
        C = 0  # normalisation constant.
        # print("LOOP",log10Pmin)
        for lP in np.arange(log10Pmin, 10, dlP):
            C += dlP * Izzard2012_period_distribution(10 ** lP, M1, log10Pmin)

        distribution_constants["Izzard2012"][M1][log10Pmin] = 1.0 / C
        # print(
        # "Normalisation constant for Izzard2012 M={} (log10Pmin={}) is\
        # {}\n".format(
        #     M1, log10Pmin, distribution_constants["Izzard2012"][M1][log10Pmin]
        # )
        # )

    lP = math.log10(P)
    # log period

    # # fits
    mu = interpolate_in_mass_izzard2012(M1, -17.8, 5.03)
    sigma = interpolate_in_mass_izzard2012(M1, 9.18, 2.28)
    K = interpolate_in_mass_izzard2012(M1, 6.93e-2, 0.0)
    nu = interpolate_in_mass_izzard2012(M1, 0.3, -1)
    g = 1.0 / (1.0 + 1e-30 ** (lP - nu))

    lPmu = lP - mu
    # print(
    #     "M={} ({}) P={} : mu={} sigma={} K={} nu={} norm=%g\n".format(
    #         Mwas, M1, P, mu, sigma, K, nu
    #     )
    # )

    # print "FUNC $distdata{Izzard2012}{$M}{$log10Pmin} * (exp(- (x-$mu)**2/(2.0*$sigma*$sigma) ) + $K/MAX(0.1,$lP)) * $g;\n";

    if (lP < log10Pmin) or (lP > 10.0):
        return 0

    else:
        return (
            distribution_constants["Izzard2012"][M1][log10Pmin]
            * (math.exp(-lPmu * lPmu / (2.0 * sigma * sigma)) + K / max(0.1, lP))
            * g
        )


########################################################################
# Mass ratio distributions
########################################################################


def flatsections(x: float, opts: dict) -> Union[float, int]:
    """
    Function to generate flat distributions, possibly in multiple sections

    Args:
        x: mass ratio value
        opts: list containing the flat sections. Which are themselves dictionaries, with keys "max": upper bound, "min": lower bound and "height": value

    Returns:
        probability of that mass ratio.
    """

    c = 0
    y = 0

    for opt in opts:
        dc = (opt["max"] - opt["min"]) * opt["height"]
        # print("added flatsection ({}-{})*{} = {}\n".format(
        #   opt['max'], opt['min'], opt['height'], dc))
        c += dc
        if opt["min"] <= x <= opt["max"]:
            y = opt["height"]
            # print("Use this\n")

    c = 1.0 / c
    y = y * c

    # print("flatsections gives C={}: y={}\n",c,y)
    return y


# print(flatsections(1, [{'min': 0, 'max': 2, 'height': 3}]))

########################################################################
# Eccentricity distributions
########################################################################

########################################################################
# Star formation histories
########################################################################


def cosmic_SFH_madau_dickinson2014(z):
    """
    Cosmic star formation history distribution from Madau & Dickonson 2014 (https://arxiv.org/pdf/1403.0007.pdf)

    Args:
        z: redshift

    Returns:
        Cosmic star formation rate in Solar mass year^-1 mega parsec^-3
    """

    CSFH = 0.015 * ((1 + z) ** 2.7) / (1 + (((1 + z) / 2.9) ** 5.6))

    return CSFH


########################################################################
# Metallicity distributions
########################################################################


########################################################################
# Moe & DiStefano 2017 functions
#
# The code below are functions that are used to set up and interpolate
# on the Moe & DiStefano 2017 data. The interpolators take the last
# known value if we try to interpolate outside of the tables.
# There are still some open tasks and improvements that can be made:
#
# TODO: Solve the memory issues that are present.
#    Are the interpolators not cleaned?
# TODO: Parallelize the setting up of the interpolators
# TODO: Generalise the code such that we can input other/newer tables

########################################################################

import py_rinterpolate

# Global dictionary to store values in
Moecache = {}


def poisson(lambda_val, n, nmax=None, verbosity=0):
    """
    Function that calculates the Poisson value and normalises
    TODO: improve the description
    """

    cachekey = "{} {} {}".format(lambda_val, n, nmax)

    if distribution_constants.get("poisson_cache", None):
        if distribution_constants["poisson_cache"].get(cachekey, None):
            p_val = distribution_constants["poisson_cache"][cachekey]

            verbose_print(
                "\tMoe and di Stefano 2017: found cached value for poisson({}, {}, {}): {}".format(
                    lambda_val, n, nmax, p_val
                ),
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )

            return p_val

    # Poisson distribution : note, n can be zero
    #
    # nmax is the truncation : if set, we normalise
    # correctly.
    p_val = _poisson(lambda_val, n)

    if nmax:
        I_poisson = 0
        for i in range(nmax + 1):
            I_poisson += _poisson(lambda_val, i)
        p_val = p_val / I_poisson

    # Add to cache
    if not distribution_constants.get("poisson_cache", None):
        distribution_constants["poisson_cache"] = {}
    distribution_constants["poisson_cache"][cachekey] = p_val

    verbose_print(
        "\tMoe and di Stefano 2017: Poisson({}, {}, {}): {}".format(
            lambda_val, n, nmax, p_val
        ),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )
    return p_val


def _poisson(lambda_val, n):
    """
    Function to return the Poisson value
    """

    return (lambda_val ** n) * np.exp(-lambda_val) / (1.0 * math.factorial(n))


def get_max_multiplicity(multiplicity_array):
    """
    Function to get the maximum multiplicity
    """

    max_multiplicity = 0
    for n in range(4):
        if multiplicity_array[n] > 0:
            max_multiplicity = n + 1
    return max_multiplicity


def merge_multiplicities(result_array, max_multiplicity, verbosity=0):
    """
    Function to fold the multiplicities higher than the max_multiplicity onto the max_multiplicity

    if max_multiplicity == 1:
        All the multiplicities are folded onto multiplicity == 1. This will always total to 1
    if max_multiplicity == 2:
        The multiplicity fractions of the triple and quadruples are folded onto that of the binary multiplicity fraction
    if max_multiplicity == 3:
        The multiplicity fractions of the quadruples are folded onto that of the triples
    """

    if not max_multiplicity in range(1, 5):
        msg = "\tMoe and di Stefano 2017: merge_multiplicities: max_multiplicity has to be between 1 and 4. It is {} now".format(
            max_multiplicity
        )
        verbose_print(
            msg,
            verbosity,
            0,
        )
        raise ValueError(msg)

    # Fold multiplicities:
    verbose_print(
        "\tMoe and di Stefano 2017: merge_multiplicities: Merging multiplicities with initial array {} and max multiplicity {}".format(
            result_array, max_multiplicity
        ),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )
    for i in range(max_multiplicity, len(result_array))[::-1]:
        result_array[i - 1] += result_array[i]
        result_array[i] = 0
    verbose_print(
        "\tMoe and di Stefano 2017: merge_multiplicities: Merging multiplicities to new array {}".format(
            result_array
        ),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )

    return result_array


def normalize_dict(result_dict, verbosity=0):
    """
    Function to normalise a dictionary
    """

    sum_result = sum([result_dict[key] for key in result_dict.keys()])
    for key in result_dict.keys():
        result_dict[key] = result_dict[key] / sum_result
    return result_dict


def Moe_di_Stefano_2017_multiplicity_fractions(options, verbosity=0):
    """
    Function that creates a list of probability fractions and
    normalises and merges them according to the users choice.

    TODO: make an extrapolation functionality in this. log10(1.6e1)
    is low, we can probably go a bit further

    The default result that is returned when sampling the mass outside
    of the mass range is now the last known value

    Returns a list of multiplicity fractions for a given input of mass
    """

    # Use the global Moecache
    global Moecache

    multiplicity_modulator_array = np.array(
        options["multiplicity_modulator"]
    )  # Modulator array

    # Check for length
    if not len(multiplicity_modulator_array) == 4:
        msg = "Multiplicity modulator has to have 4 elements. Now it is {}, len: {}".format(
            multiplicity_modulator_array, len(multiplicity_modulator_array)
        )
        verbose_print(
            msg,
            verbosity,
            0,
        )
        raise ValueError(msg)

    # Set up some arrays
    full_fractions_array = np.zeros(4)  # Meant to contain the real fractions
    weighted_fractions_array = np.zeros(
        4
    )  # Meant to contain the fractions multiplied by the multiplicity modulator

    # Get max multiplicity
    max_multiplicity = get_max_multiplicity(multiplicity_modulator_array)

    # ... it's better to interpolate the multiplicity and then
    # use a Poisson distribution to calculate the fractions
    # (this is more accurate)

    # Set up the multiplicity interpolator
    if not Moecache.get("rinterpolator_multiplicity", None):
        Moecache["rinterpolator_multiplicity"] = py_rinterpolate.Rinterpolate(
            table=Moecache["multiplicity_table"],  # Contains the table of data
            nparams=1,  # logM1
            ndata=4,  # The number of datapoints (the parameters that we want to interpolate)
            verbosity=verbosity - (_MOE2017_VERBOSITY_INTERPOLATOR_LEVEL - 1),
        )

    if options["multiplicity_model"] == "Poisson":
        multiplicity = Moecache["rinterpolator_multiplicity"].interpolate(
            [np.log10(options["M_1"])]
        )[0]

        # Fill the multiplicity array
        for n in range(4):
            full_fractions_array[n] = poisson(multiplicity, n, 3, verbosity)

        # Normalize it so it fills to one when taking all the multiplicities:
        full_fractions_array = full_fractions_array / np.sum(full_fractions_array)

        verbose_print(
            "\tMoe and di Stefano 2017: Moe_di_Stefano_2017_multiplicity_fractions: using model {}: full_fractions_array: {}".format(
                "Poisson", full_fractions_array
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

    elif options["multiplicity_model"] == "data":
        # use the fractions calculated from Moe's data directly
        #
        # note that in this case, there are no quadruples: these
        # are combined with triples

        # Fill with the raw values
        for n in range(3):
            full_fractions_array[n] = Moecache[
                "rinterpolator_multiplicity"
            ].interpolate([np.log10(options["M_1"])])[n + 1]

        # Set last value
        full_fractions_array[3] = 0.0  # no quadruples
        verbose_print(
            "\tMoe and di Stefano 2017: Moe_di_Stefano_2017_multiplicity_fractions: using model {}: full_fractions_array: {}".format(
                "data", full_fractions_array
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

    # Normalisation:
    if options["normalize_multiplicities"] == "raw":
        # Don't multiply by the multiplicity_array, but do give a fractions array
        verbose_print(
            "\tMoe and di Stefano 2017: Moe_di_Stefano_2017_multiplicity_fractions: Not normalising (using raw results): results: {}".format(
                full_fractions_array
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
        result = full_fractions_array

    elif options["normalize_multiplicities"] == "norm":
        # Multiply the full_multiplicity_fraction array by the multiplicity_multiplier_array, creating a weighted fractions array
        weighted_fractions_array = full_fractions_array * multiplicity_modulator_array

        # Normalise this so it is in total 1:
        result = weighted_fractions_array / np.sum(weighted_fractions_array)

        verbose_print(
            "\tMoe and di Stefano 2017: Moe_di_Stefano_2017_multiplicity_fractions: Normalising with {}. result: {}".format(
                "norm", result
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

    elif options["normalize_multiplicities"] == "merge":
        # We first take the full multiplicity array
        # (i.e. not multiplied by multiplier) and do the merging
        result = merge_multiplicities(
            full_fractions_array, max_multiplicity, verbosity=verbosity
        )

        # Then normalise to be sure
        result = result / np.sum(result)

        verbose_print(
            "\tMoe and di Stefano 2017: Moe_di_Stefano_2017_multiplicity_fractions: Normalising with {}, max_multiplicity={} result={}".format(
                "merge", max_multiplicity, result
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

    verbose_print(
        "\tMoe and di Stefano 2017: Moe_di_Stefano_2017_multiplicity_fractions: {}".format(
            str(result)
        ),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )

    # return array reference
    return result


def build_q_table(options, m, p, verbosity=0):
    ############################################################
    #
    # Build an interpolation table for q, given a mass and
    # orbital period.
    #
    # $m and $p are labels which determine which system(s)
    # to look up from Moe's data:
    #
    # $m can be M1, M2, M3, M4, or if set M1+M2 etc.
    # $p can be P, P2, P3
    #
    # The actual values are in $opts:
    #
    # mass is in $opts->{$m}
    # period is  $opts->{$p}
    #
    # Since the information from the table for Moe and di Stefano 2017 is independent of any choice we make,
    # we need to take into account that for example our choice of minimum mass leads to
    # a minimum q_min that is not the same as in the table
    # We should ignore those parts of the table and renormalise.
    # If we are below the lowest value of qmin in the table we need to extrapolate the data
    #
    # Anyway, the goal of this function is to provide some extrapolated values for q when we should sample outside of the boundaries
    ############################################################

    # We can check if we have a cached value for this already:
    # TODO: fix this cache check.
    incache = False
    if Moecache.get("rinterpolator_q_metadata", None):
        if (Moecache["rinterpolator_q_metadata"].get(m, None)) and (
            Moecache["rinterpolator_q_metadata"].get(p, None)
        ):
            if (Moecache["rinterpolator_q_metadata"][m] == options[m]) and (
                Moecache["rinterpolator_q_metadata"][p] == options[p]
            ):
                incache = True

                verbose_print(
                    "\tMoe and di Stefano 2017: build_q_table: Found cached values for m={} p={}".format(
                        options[m], options[p]
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )
            else:
                verbose_print(
                    "\tMoe and di Stefano 2017: build_q_table: Cached values for different m={} p={}. Freeing current table and making new table".format(
                        options[m], options[p]
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )

    #
    if not incache:
        # trim and/or expand the table to the range $qmin to $qmax.

        # qmin is set by the minimum stellar mass : below this
        # the companions are planets
        # qmin = options["ranges"]["M"][
        #     0
        # ]  # TODO: this lower range must not be lower than Mmin.

        qmin = options["Mmin"] / options["M_1"]
        verbose_print(
            "\tMoe and di Stefano 2017: build_q_table qmin: {}".format(
                qmin,
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

        # qmax = maximum_mass_ratio_for_RLOF(options[m], options[p])
        # TODO: change this to the above
        qmax = 1

        # qdata contains the table that we modify: we get
        # the original data by interpolating Moe's table
        qdata = {}
        can_renormalize = 1

        qeps = 1e-8  # small number but such that qeps+1 != 1
        if qeps + 1 == 1.0:
            verbose_print(
                "\tMoe and di Stefano 2017: build_q_table: qeps (= {}) +1 == 1. Make qeps larger".format(
                    qeps,
                ),
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )

        if qmin >= qmax:
            # there may be NO binaries in this part of the parameter space:
            # in which case, set up a table with lots of zero in it

            qdata = {0: 0, 1: 0}
            can_renormalize = 0

        else:
            # qmin and qmax mean we'll get something non-zero
            can_renormalize = 1

            # require extrapolation sets whether we need to extrapolate
            # at the low and high ends
            require_extrapolation = {}

            if qmin >= 0.15:
                # qmin is inside Moe's table : this is easy,
                # we just keep points from qmin at the low
                # end to qmax at the high end.
                require_extrapolation["low"] = 0
                require_extrapolation[
                    "high"
                ] = 1  # TODO: shouldn't the extrapolation need to happen if qmax > 0.95
                qdata[qmin] = Moecache["rinterpolator_q"].interpolate(
                    [np.log10(options[m]), np.log10(options[p]), qmin]
                )[0]

                for q in np.arange(0.15, 0.950001, 0.1):
                    if (q >= qmin) and (q <= qmax):
                        qdata[q] = Moecache["rinterpolator_q"].interpolate(
                            [np.log10(options[m]), np.log10(options[p]), q]
                        )[0]
            else:
                require_extrapolation["low"] = 1
                require_extrapolation["high"] = 1
                if qmax < 0.15:
                    # qmax < 0.15 which is off the edge
                    # of the table. In this case, choose
                    # two points at q=0.15 and 0.16 and interpolate
                    # at these in case we want to extrapolate.
                    for q in [0.15, 0.16]:
                        qdata[q] = Moecache["rinterpolator_q"].interpolate(
                            [np.log10(options[m]), np.log10(options[p]), q]
                        )[0]
                else:
                    # qmin < 0.15 and qmax > 0.15, so we
                    # have to generate Moe's table for
                    # q = 0.15 (i.e. 0.1 to 0.2) to 0.95 (0.9 to 1)
                    # as a function of M1 and orbital period,
                    # to obtain the q distribution data.

                    for q in np.arange(0.15, np.min([0.950001, qmax + 0.0001]), 0.1):
                        val = Moecache["rinterpolator_q"].interpolate(
                            [np.log10(options[m]), np.log10(options[p]), q]
                        )[0]
                        qdata[q] = val

                # just below qmin, if qmin>qeps, we want nothing
                if qmin - 0.15 > qeps:
                    q = qmin - qeps
                    qdata[q] = 0
                    require_extrapolation["low"] = 0

            # just above qmax, if qmax<1, we want nothing
            if qmax < 0.95:
                q = qmax + qeps
                qdata[q] = 0
                require_extrapolation["high"] = 0

            # sorted list of qs
            qs = sorted(qdata.keys())

            if len(qs) == 0:
                msg = "No qs found error"
                raise ValueError(msg)

            elif len(qs) == 1:
                # only one q value : pretend there are two
                # with a flat distribution up to 1.0.
                if qs[0] == 1.0:
                    qs[0] = 1.0 - 1e-6
                    qs.append(1)
                    qdata[qs[0]] = 1
                    qdata[qs[1]] = 1
                else:
                    qs.append(1)
                    qdata[qs[1]] = qs[0]

            # We actually should do the extrapolation now.
            else:
                # Loop over both the lower end and the upper end
                for pre in ["low", "high"]:
                    if require_extrapolation[pre] == 0:
                        continue
                    else:
                        sign = -1 if pre == "low" else 1
                        end_index = 0 if pre == "low" else len(qs) - 1
                        indices = (
                            [0, 1] if pre == "low" else [len(qs) - 1, len(qs) - 2]
                        )  # Based on whether we do the high or low end we need to use two different indices
                        method = options.get(
                            "q_{}_extrapolation_method".format(pre), None
                        )
                        qlimit = qmin if pre == "low" else qmax

                        verbose_print(
                            "\tMoe and di Stefano 2017: build_q_table: Extrapolating: Q: {} method: {}, indices: {} End index: {}".format(
                                pre, method, indices, end_index
                            ),
                            verbosity,
                            _MOE2017_VERBOSITY_LEVEL,
                        )

                        # truncate the distribution
                        qdata[max(0.0, min(1.0, qlimit + sign * qeps))] = 0

                        if method == None:
                            # no extrapolation : just interpolate between 0.10 and 0.95
                            verbose_print(
                                "\tMoe and di Stefano 2017: build_q_table: using no extrapolations".format(),
                                verbosity,
                                _MOE2017_VERBOSITY_LEVEL,
                            )
                            continue
                        elif method == "flat":
                            # use the end value and extrapolate it
                            # with zero slope
                            qdata[qlimit] = qdata[qs[end_index]]
                            verbose_print(
                                "\tMoe and di Stefano 2017: build_q_table: using constant extrapolation".format(),
                                verbosity,
                                _MOE2017_VERBOSITY_LEVEL,
                            )
                        elif method == "linear":
                            qdata[qlimit] = linear_extrapolation_q(
                                qs=qs,
                                indices=indices,
                                qlimit=qlimit,
                                qdata=qdata,
                                end_index=end_index,
                                verbosity=verbosity,
                            )

                            verbose_print(
                                "\tMoe and di Stefano 2017: build_q_table: using linear extrapolation".format(),
                                verbosity,
                                _MOE2017_VERBOSITY_LEVEL,
                            )
                            if pre == "low":
                                below_qlimit = qlimit - qeps
                                if below_qlimit > 0:
                                    qdata[below_qlimit] = 0
                                qdata[0] = 0
                                verbose_print(
                                    "\tMoe and di Stefano 2017: build_q_table: using linear extrapolation and setting the points below the lower q bound ({}) to 0 ".format(
                                        qlimit
                                    ),
                                    verbosity,
                                    _MOE2017_VERBOSITY_LEVEL,
                                )

                        elif method == "plaw2":
                            qdata[qlimit] = powerlaw_extrapolation_q(
                                qs=qs, indices=indices, qdata=qdata, verbosity=verbosity
                            )

                            verbose_print(
                                "\tMoe and di Stefano 2017: build_q_table: using powerlaw extrapolation".format(),
                                verbosity,
                                _MOE2017_VERBOSITY_LEVEL,
                            )
                        elif method == "nolowq":
                            newq = 0.05
                            qdata[newq] = 0
                            verbose_print(
                                "\tMoe and di Stefano 2017: build_q_table: setting lowq to 0".format(),
                                verbosity,
                                _MOE2017_VERBOSITY_LEVEL,
                            )
                        elif method == "poly":
                            # TODO: consider implementing the poly method (see Perl version)
                            raise ValueError(
                                "Moe and di Stefano 2017: build_q_table: Method 'poly' not implemented"
                            )

                        else:
                            msg = "\tMoe and di Stefano 2017: build_q_table: Error no other methods available. The chosen method ({}) does not exist!".format(
                                method
                            )
                            verbose_print(
                                msg,
                                verbosity,
                                _MOE2017_VERBOSITY_LEVEL,
                            )
                            raise ValueError(msg)

        # regenerate qs in new table. This is now the updated list of qs where we have some extrapolated numbers
        tmp_table = []
        for q in sorted(qdata.keys()):
            tmp_table.append([q, qdata[q]])

        # Make an interpolation table to contain our modified data
        q_interpolator = py_rinterpolate.Rinterpolate(
            table=tmp_table,
            nparams=1,
            ndata=1,  # Contains the table of data  # q  #
            verbosity=verbosity - (_MOE2017_VERBOSITY_INTERPOLATOR_LEVEL - 1),
        )
        verbose_print(
            "\tMoe and di Stefano 2017: build_q_table: Created a new Q table",
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

        if can_renormalize:
            verbose_print(
                "\tMoe and di Stefano 2017: build_q_table: Renormalising table",
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )

            # now we integrate and renormalise (if the table is not all zero)
            I = get_integration_constant_q(
                q_interpolator, tmp_table, qdata, verbosity=verbosity
            )

            if I > 0:
                # normalise to 1.0 by dividing the data by 1.0/$I
                q_interpolator.multiply_table_column(1, 1.0 / I)

                # test this
                new_I = get_integration_constant_q(
                    q_interpolator, tmp_table, qdata, verbosity=verbosity
                )

                # fail if error in integral > 1e-6 (should be ~ machine precision)
                if abs(1.0 - new_I) > 1e-6:
                    verbose_print(
                        "\tMoe and di Stefano 2017: build_q_table: Error: > 1e-6 in q probability integral: {}".format(
                            I
                        ),
                        verbosity,
                        _MOE2017_VERBOSITY_LEVEL,
                    )
        # set this new table in the cache
        Moecache["rinterpolator_q_given_{}_log10{}".format(m, p)] = q_interpolator
        verbose_print(
            "\tMoe and di Stefano 2017: build_q_table: stored q_interpolater as {}".format(
                "rinterpolator_q_given_{}_log10{}".format(m, p)
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )

        # Store the values for which this table was set up in the dict
        if not Moecache.get("rinterpolator_q_metadata", None):
            Moecache["rinterpolator_q_metadata"] = {}
        Moecache["rinterpolator_q_metadata"][m] = options[m]
        Moecache["rinterpolator_q_metadata"][p] = options[p]


def powerlaw_extrapolation_q(qdata, qs, indices, verbosity=0):
    """
    Function to do the power law extrapolation at the lower end of the q range
    """
    newq = 0.05

    # use a power-law extrapolation down to q=0.05, if possible
    if (qdata[qs[indices[0]]] == 0.0) and (qdata[qs[indices[1]]] == 0.0):
        # not possible
        return 0

    else:
        slope = (np.log10(qdata[qs[indices[1]]]) - np.log10(qdata[qs[indices[0]]])) / (
            np.log10(qs[indices[1]]) - np.log10(qs[indices[0]])
        )
        intercept = np.log10(qdata[qs[indices[0]]]) - slope * np.log10(qs[indices[0]])

        return slope * newq + intercept


def linear_extrapolation_q(qs, indices, qlimit, qdata, end_index, verbosity=0):
    """
    Function to do the linear extrapolation for q.
    """

    # linear extrapolation
    dq = qs[indices[1]] - qs[indices[0]]

    if dq == 0:
        verbose_print(
            "\tMoe and di Stefano 2017: build_q_table: linear dq=0".format(),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
        # No change
        return qs[end_index]
    else:
        slope = (qdata[qs[indices[1]]] - qdata[qs[indices[0]]]) / dq

        intercept = qdata[qs[indices[0]]] - slope * qs[indices[0]]
        qdata[qlimit] = max(0.0, slope * qlimit + intercept)
        verbose_print(
            "\tMoe and di Stefano 2017: build_q_table: linear Slope: {} intercept: {} dn/dq({}) = {}".format(
                slope, intercept, qlimit, qdata[qlimit]
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    return max(0.0, slope * qlimit + intercept)


def get_integration_constant_q(q_interpolator, tmp_table, qdata, verbosity=0):
    """
    Function to integrate the q interpolator and return the integration constant
    """

    dq = 1e-3  # resolution of the integration/renormalisation
    I = 0

    # integrate: note that the value of the integral is
    # meaningless to within a factor (which depends on $dq)
    for q in np.arange(0, 1 + 2e-6, dq):
        x = q_interpolator.interpolate([q])
        if len(x) == 0:
            msg = "\tMoe and di Stefano 2017: build_q_table: Q interpolator table interpolation failed.\n\t\ttmp_table = {}\n\t\tq_data = {}".format(
                str(tmp_table), str(qdata)
            )
            verbose_print(
                msg,
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )
            raise ValueError(msg)
        else:
            I += x[0] * dq
            # verbose_print(
            #     "\tMoe and di Stefano 2017: build_q_table: dn/dq ({}) = {} I -> = {}".format(q, x[0], I),
            #     verbosity,
            #     _MOE2017_VERBOSITY_LEVEL,
            # )
    return I


def fill_data(sample_values, data_dict):
    """
    Function that returns the normalised array of values for given logmass and logperiod
    used for the e and q values

    TODO: make sure we do the correct thing with the dstep
    """

    data = {}
    I = 0

    dstep = float(sample_values[1]) - float(sample_values[0])

    # Read out the data
    for sample_value in sample_values:
        val = data_dict[sample_value]
        data[sample_value] = val
        I += val

    # Normalise the data
    for sample_value in sample_values:
        data[sample_value] = data[sample_value] / I

    return data


def calc_e_integral(
    options,
    integrals_string,
    interpolator_name,
    mass_string,
    period_string,
    verbosity=0,
):
    """
    Function to calculate the P integral

    We need to renormalise this because min_per > 0, and not all periods should be included
    """

    global Moecache
    min_ecc = 0
    max_ecc = 0.9999

    mass_period_string = "{}_{}".format(options[mass_string], options[period_string])

    # Check if the dict exists
    if not Moecache.get(integrals_string, None):
        Moecache[integrals_string] = {}

    # Check for cached value. If it doesn't exist: calculate
    if not Moecache[integrals_string].get(mass_period_string, None):
        I = 0
        decc = 1e-3

        for ecc in np.arange(min_ecc, max_ecc, decc):
            # Loop over all the values in the table, between the min and max P
            dp_decc = Moecache[interpolator_name].interpolate(
                [np.log10(options[mass_string]), np.log10(options[period_string]), ecc]
            )[0]

            I += dp_decc * decc

        # Set the integral value in the dict
        Moecache[integrals_string][mass_period_string] = I
        verbose_print(
            "\tMoe and di Stefano 2017: calc_ecc_integral: min_ecc: {} max ecc: {} integrals_string: {} interpolator_name: {} mass_string: {} period_string: {} mass: {} period: {} I: {}".format(
                min_ecc,
                max_ecc,
                integrals_string,
                interpolator_name,
                mass_string,
                period_string,
                options[mass_string],
                options[period_string],
                I,
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    else:
        verbose_print(
            "\tMoe and di Stefano 2017: calc_ecc_integral: Found cached value for min_ecc: {} max ecc: {} integrals_string: {} interpolator_name: {} mass_string: {} period_string: {} mass: {} period: {} I: {}".format(
                min_ecc,
                max_ecc,
                integrals_string,
                interpolator_name,
                mass_string,
                period_string,
                options[mass_string],
                options[period_string],
                Moecache[integrals_string][mass_period_string],
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )


def calc_P_integral(
    options,
    min_logP,
    max_logP,
    integrals_string,
    interpolator_name,
    mass_string,
    verbosity=0,
):
    """
    Function to calculate the P integral

    We need to renormalise this because min_per > 0, and not all periods should be included
    """

    global Moecache

    # Check if the dict exists
    if not Moecache.get(integrals_string, None):
        Moecache[integrals_string] = {}

    # Check for cached value. If it doesn't exist: calculate
    if not Moecache[integrals_string].get(options[mass_string], None):
        I = 0
        dlogP = 1e-3

        for logP in np.arange(min_logP, max_logP, dlogP):
            # Loop over all the values in the table, between the min and max P
            dp_dlogP = Moecache[interpolator_name].interpolate(
                [np.log10(options[mass_string]), logP]
            )[0]

            I += dp_dlogP * dlogP

        # Set the integral value in the dict
        Moecache[integrals_string][options[mass_string]] = I
        verbose_print(
            "\tMoe and di Stefano 2017: calc_P_integral: min_logP: {} integrals_string: {} interpolator_name: {} mass_string: {} mass: {} I: {}".format(
                min_logP,
                integrals_string,
                interpolator_name,
                mass_string,
                options[mass_string],
                I,
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    else:
        verbose_print(
            "\tMoe and di Stefano 2017: calc_P_integral: Found cached value for min_logP: {} integrals_string: {} interpolator_name: {} mass_string: {} mass: {} I: {}".format(
                min_logP,
                integrals_string,
                interpolator_name,
                mass_string,
                options[mass_string],
                Moecache[integrals_string][options[mass_string]],
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )


def calc_total_probdens(prob_dict):
    """
    Function to calculate the total probability density
    """

    total_probdens = 1
    for key in prob_dict:
        total_probdens *= prob_dict[key]
    prob_dict["total_probdens"] = total_probdens

    return prob_dict


def Moe_di_Stefano_2017_pdf(options, verbosity=0):
    """
    Moe & diStefano function to calculate the probability density.

    takes a dictionary as input (in options) with options:

    M1, M2, M3, M4 => masses (Msun) [M1 required, rest optional]
    P, P2, P3 => periods (days) [number: none=binary, 2=triple, 3=quadruple]
    ecc, ecc2, ecc3 => eccentricities [numbering as for P above]

    mmin => minimum allowed stellar mass (default 0.07)
    mmax => maximum allowed stellar mass (default 80.0)
    """

    verbose_print(
        "\tMoe_di_Stefano_2017_pdf with options:\n\t\t{}".format(json.dumps(options)),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )

    prob_dict = (
        {}
    )  # Dictionary containing all the pdf values for the different parameters

    # Get the multiplicity from the options, and if its not there, calculate it based on the
    # TODO: the function below makes no sense. We NEED to pass the multiplicity in the
    if not options.get("multiplicity", None):
        msg = "\tMoe_di_Stefano_2017_pdf: Did not find a multiplicity value in the options dictionary"
        verbose_print(
            msg,
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
        raise ValueError(msg)
        # multiplicity = 1
        # for n in range(2, 5):
        #     multiplicity += 1 if options.get("M{}".format(n), None) else 0
    else:
        multiplicity = options["multiplicity"]

    # Immediately return 0 if the multiplicity modulator is 0
    if options["multiplicity_modulator"][multiplicity - 1] == 0:
        verbose_print(
            "\tMoe_di_Stefano_2017_pdf: returning 0 because of the multiplicity modulator being 0",
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
        return 0

    ############################################################
    # multiplicity fraction
    # Calculate the probability, or rather, fraction, of stars that belong to this mass

    multiplicity_probability = Moe_di_Stefano_2017_multiplicity_fractions(
        options, verbosity
    )[multiplicity - 1]
    prob_dict["multiplicity"] = multiplicity_probability
    verbose_print(
        "\tMoe_di_Stefano_2017_pdf: Appended multiplicity (mass1 = {}) probability ({}) to the prob dict ({})".format(
            options["M_1"], prob_dict["multiplicity"], prob_dict
        ),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )

    ############################################################
    # always require an IMF for the primary star
    #
    # NB multiply by M1 to convert dN/dM to dN/dlnM
    # (dlnM = dM/M, so 1/dlnM = M/dM)

    # TODO: Create an n-part-powerlaw method that can have breakpoints and slopes. I'm using a three-part power law now.
    # TODO: is this actually the correct way? putting the M1 in there? Do we sample in log space?
    M1_probability = Kroupa2001(options["M_1"]) * options["M_1"]
    prob_dict["M_1"] = M1_probability
    verbose_print(
        "\tMoe_di_Stefano_2017_pdf: Appended Mass (m={}) probability ({}) to the prob dict ({})".format(
            options["M_1"], prob_dict["M_1"], prob_dict
        ),
        verbosity,
        _MOE2017_VERBOSITY_LEVEL,
    )
    # if M1_probability == 0: # If the probability is 0 then we don't have to calculate more
    #     calc_total_probdens(prob_dict)
    #     return prob_dict

    """
    From here we go through the multiplicities.
    """
    if multiplicity >= 2:
        # If the multiplicity is higher than 1, we will need to construct the following tables:
        # - period distribution table
        # - q distribution table
        # - eccentricity distribution table

        # Set up the interpolator for the periods
        if not Moecache.get("rinterpolator_log10P", None):
            Moecache["rinterpolator_log10P"] = py_rinterpolate.Rinterpolate(
                table=Moecache["period_distributions"],  # Contains the table of data
                nparams=2,  # log10M, log10P
                ndata=2,  # binary, triple
                verbosity=verbosity - (_MOE2017_VERBOSITY_INTERPOLATOR_LEVEL - 1),
            )
            verbose_print(
                "\tMoe_di_Stefano_2017_pdf: Created new period interpolator: {}".format(
                    Moecache["rinterpolator_log10P"]
                ),
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )

        # Make a table storing Moe's data for q distributions
        if (
            options.get("M_2", None)
            or options.get("M_3", None)
            or options.get("M_4", None)
        ):
            if not Moecache.get("rinterpolator_q", None):
                Moecache["rinterpolator_q"] = py_rinterpolate.Rinterpolate(
                    table=Moecache["q_distributions"],  # Contains the table of data
                    nparams=3,  # log10M, log10P, q
                    ndata=1,  #
                    verbosity=verbosity - (_MOE2017_VERBOSITY_INTERPOLATOR_LEVEL - 1),
                )
                verbose_print(
                    "\tMoe_di_Stefano_2017_pdf: Created new q interpolator: {}".format(
                        Moecache["rinterpolator_q"]
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )

        # Make a table storing Moe's data for q distributions, but only if the ecc is actually sampled
        if "ecc" in options:
            if not options["ecc"] == None:
                if not Moecache.get("rinterpolator_e", None):
                    Moecache["rinterpolator_e"] = py_rinterpolate.Rinterpolate(
                        table=Moecache[
                            "ecc_distributions"
                        ],  # Contains the table of data
                        nparams=3,  # log10M, log10P, e
                        ndata=1,  #
                        verbosity=verbosity
                        - (_MOE2017_VERBOSITY_INTERPOLATOR_LEVEL - 1),
                    )
                    verbose_print(
                        "\tMoe_di_Stefano_2017_pdf: Created new e interpolator: {}".format(
                            Moecache["rinterpolator_e"]
                        ),
                        verbosity,
                        _MOE2017_VERBOSITY_LEVEL,
                    )

        ###############
        # Calculation for period of the binary

        if options.get("M_2", None):
            # Separation of the inner binary
            options["sep"] = calc_sep_from_period(
                options["M_1"], options["M_2"], options["P"]
            )
            # TODO: add check for min_logP with instant RLOF?
            # TODO: Actually use the value above.
            # Total mass inner binary:
            options["M_1+M_2"] = options["M_1"] + options["M_2"]

        # Calculate P integral or use cached value

        # get the periods from the Moecahe
        min_logP = float(Moecache["logperiods"][0])
        max_logP = float(Moecache["logperiods"][-1])

        calc_P_integral(
            options,
            min_logP,
            max_logP,
            "P_integrals",
            "rinterpolator_log10P",
            "M_1",
            verbosity,
        )

        # Set probabilty for P1
        p_val = Moecache["rinterpolator_log10P"].interpolate(
            [np.log10(options["M_1"]), np.log10(options["P"])]
        )[0]
        p_val = p_val / Moecache["P_integrals"][options["M_1"]]
        prob_dict["P"] = p_val
        verbose_print(
            "\tMoe_di_Stefano_2017_pdf: Appended period (m={}, P={}) probability ({}) to the prob list ({})".format(
                options["M_1"], options["P"], prob_dict["P"], prob_dict
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
        # if prob_dict['P'] == 0: # If the probability is 0 then we don't have to calculate more
        #     calc_total_probdens(prob_dict)
        #     return prob_dict

        ############################################################
        # mass ratio (0 < q = M2/M1 < qmax)
        #
        # we need to construct the q table for the given M1
        # subject to qmin = Mmin/M1

        if options.get("M_2", None):
            # Build the table for q
            primary_mass = options["M_1"]
            secondary_mass = options["M_2"]
            m_label = "M_1"
            p_label = "P"

            # Construct the q table
            build_q_table(options, m_label, p_label, verbosity=verbosity)
            verbose_print(
                "\tMoe_di_Stefano_2017_pdf: Created q_table ({}) for m={} p={}".format(
                    Moecache[
                        "rinterpolator_q_given_{}_log10{}".format(m_label, p_label)
                    ],
                    options[m_label],
                    options[p_label],
                ),
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )

            # Add probability for the mass ratio
            q_prob = Moecache[
                "rinterpolator_q_given_{}_log10{}".format(m_label, p_label)
            ].interpolate([secondary_mass / primary_mass])[0]
            prob_dict["q"] = q_prob
            verbose_print(
                "\tMoe_di_Stefano_2017_pdf: appended mass ratio (M={} P={} q={}) probability ({}) to the prob list ({}) ".format(
                    options["M_1"],
                    options["P"],
                    options["M_2"] / options["M_1"],
                    prob_dict["q"],
                    prob_dict,
                ),
                verbosity,
                _MOE2017_VERBOSITY_LEVEL,
            )
            # if prob_dict['q'] == 0: # If the probability is 0 then we don't have to calculate more
            #     calc_total_probdens(prob_dict)
            #     return prob_dict

        ############################################################
        # Eccentricity
        # TODO: ask rob if the eccentricity requires an extrapolation as well.

        # Only do this if the eccentricity is sampled
        if "ecc" in options:
            if not options["ecc"] == None:
                # Calculate ecc integral or use cached value
                calc_e_integral(
                    options, "ecc_integrals", "rinterpolator_e", "M_1", "P", verbosity
                )
                mass_period_string = "{}_{}".format(options["M_1"], options["P"])

                # Set probability for ecc
                ecc_val = Moecache["rinterpolator_e"].interpolate(
                    [np.log10(options["M_1"]), np.log10(options["P"]), options["ecc"]]
                )[0]
                ecc_val = ecc_val / Moecache["ecc_integrals"][mass_period_string]
                prob_dict["ecc"] = ecc_val
                verbose_print(
                    "\tMoe_di_Stefano_2017_pdf: Appended eccentricity (m={}, P={}, ecc={}) probability ({}) to the prob list ({})".format(
                        options["M_1"],
                        options["P"],
                        options["ecc"],
                        prob_dict["ecc"],
                        prob_dict,
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )
                # if prob_dict['ecc'] == 0: # If the probability is 0 then we don't have to calculate more
                #     calc_total_probdens(prob_dict)
                #     return prob_dict

        # Calculations for when multiplicity is bigger than 3
        # BEWARE: binary_c does not evolve these systems actually and the code below should be revised for when binary_c actually evolves triples.
        # For that reason, I would not advise to use things with multiplicity > 3
        if multiplicity >= 3:

            ############################################################
            # orbital period 2 =
            #     orbital period of star 3 (multiplicity==3) or
            #     the star3+star4 binary (multiplicity==4)
            #
            # we assume the same period distribution for star 3
            # (or stars 3 and 4) but with a separation that is >10*a*(1+e)
            # where 10*a*(1+e) is the maximum apastron separation of
            # stars 1 and 2

            # TODO: Is this a correct assumption?
            max_sep = 10.0 * options["sep"] * (1.0 + options["ecc"])
            min_P2 = calc_period_from_sep(options["M_1+M_2"], options["mmin"], max_sep)
            min_logP2 = math.log10(min_P2)
            # max_logP2 = 10.0
            # min_logP = Moecache['logperiods'][0]
            max_logP2 = float(Moecache["logperiods"][-1])

            if options["P2"] < min_P2:
                # period is too short : system is not hierarchical
                prob_dict["P2"] = 0
                verbose_print(
                    "\tMoe_di_Stefano_2017_pdf: period2 is too short: {} < {}, system is not hierarchical. Added 0 to probability list".format(
                        options["P1"], min_P2
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )
                # if prob_dict['P2'] == 0: # If the probability is 0 then we don't have to calculate more
                #     calc_total_probdens(prob_dict)
                #     return prob_dict

            else:
                # period is long enough that the system is hierarchical
                # hence the separation between the outer star
                # and inner binary
                options["sep2"] = calc_sep_from_period(
                    options["M_3"], options["M_1+M_2"], options["P2"]
                )

                # Check for cached value of P integral or calculate
                calc_P_integral(
                    options,
                    min_logP2,
                    max_logP2,
                    "P2_integrals",
                    "rinterpolator_log10P",
                    "M_1+M_2",
                    verbosity,
                )

                # Add the probability
                p_val = Moecache["rinterpolator_log10P"].interpolate(
                    [np.log10(options["M_1+M_2"]), np.log10(options["P2"])]
                )[0]
                p_val = p_val / Moecache["P2_integrals"][options["M_1+M_2"]]
                prob_dict["P2"] = p_val
                verbose_print(
                    "\tMoe_di_Stefano_2017_pdf: Appended period2 (m1={} m2={}, P2={}) probability ({}) to the prob list ({})".format(
                        options["M_1"],
                        options["M_2"],
                        options["P2"],
                        prob_dict["P2"],
                        prob_dict,
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )
                # if prob_dict['P2'] == 0: # If the probability is 0 then we don't have to calculate more
                #     calc_total_probdens(prob_dict)
                #     return prob_dict

                ############################################################
                # mass ratio 2 = q2 = M3 / (M1+M2)
                #
                # we need to construct the q table for the given M1
                # subject to qmin = Mmin/(M1+M2)

                # Set the variables for the masses and their names
                primary_mass = options["M_1+M_2"]
                secondary_mass = options["M_3"]
                m_label = "M_1+M_2"
                p_label = "P2"

                # Build q table
                build_q_table(options, m_label, p_label, verbosity=verbosity)
                verbose_print(
                    "\tMoe_di_Stefano_2017_pdf: Called build_q_table",
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )

                # Add the probability
                q2_val = Moecache[
                    "rinterpolator_q_given_{}_log10{}".format(m_label, p_label)
                ].interpolate([secondary_mass / primary_mass])[0]
                prob_dict["q2"] = q2_val
                verbose_print(
                    "\tMoe_di_Stefano_2017_pdf: appended mass ratio (M_1+M_2={} M_3={} P={} q={}) probability ({}) to the prob list ({}) ".format(
                        options["M_1+M_2"],
                        options["M_3"],
                        options["P"],
                        secondary_mass / primary_mass,
                        prob_dict["q2"],
                        prob_dict,
                    ),
                    verbosity,
                    _MOE2017_VERBOSITY_LEVEL,
                )
                # if prob_dict['q2'] == 0: # If the probability is 0 then we don't have to calculate more
                #     calc_total_probdens(prob_dict)
                #     return prob_dict

                # TODO: Implement ecc2 calculation
                if multiplicity == 4:
                    # quadruple system.
                    # TODO: Ask Rob about the structure of the quadruple. Is this only double binary quadruples?

                    ############################################################
                    # orbital period 3
                    #
                    # we assume the same period distribution for star 4
                    # as for any other stars but Pmax must be such that
                    # sep3 < sep2 * 0.2

                    # TODO: fix this here
                    max_sep3 = 0.2 * options["sep2"] * (1.0 + options["ecc2"])
                    max_per3 = calc_period_from_sep(
                        options["M_1+M_2"], options["mmin"], max_sep3
                    )

                    # Calculate P integral or use the cached value
                    # TODO: Make sure we use the correct period idea here.
                    calc_P_integral(
                        options,
                        min_logP2,
                        max_logP2,
                        "P2_integrals",
                        "rinterpolator_log10P",
                        "M_1+M_2",
                        verbosity,
                    )

                    # Set probability
                    p_val = Moecache["rinterpolator_log10P"].interpolate(
                        [np.log10(options["M_1+M_2"]), np.log10(options["P2"])]
                    )[0]
                    p_val = p_val / Moecache["P2_integrals"][options["M_1+M_2"]]
                    prob_dict["P3"] = p_val
                    verbose_print(
                        "\tMoe_di_Stefano_2017_pdf: Appended period2 (M=4) (M_1={} M_2={}, P2={}) probability ({}) to the prob list ({})".format(
                            options["M_1"],
                            options["M_2"],
                            options["P2"],
                            prob_dict["P3"],
                            prob_dict,
                        ),
                        verbosity,
                        _MOE2017_VERBOSITY_LEVEL,
                    )
                    # if prob_dict['P3'] == 0: # If the probability is 0 then we don't have to calculate more
                    #     calc_total_probdens(prob_dict)
                    #     return prob_dict

                    ############################################################
                    # mass ratio 2
                    #
                    # we need to construct the q table for the given M1
                    # subject to qmin = Mmin/(M1+M2)
                    # Make a table storing Moe's data for q distributions

                    # Build the table for q2
                    primary_mass = options["M_1+M_2"]
                    secondary_mass = options["M_3"]
                    m_label = "M_1+M_2"
                    p_label = "P2"

                    # Calculate new q table
                    build_q_table(options, m_label, p_label, verbosity=verbosity)
                    verbose_print(
                        "\tMoe_di_Stefano_2017_pdf: Created q_table ".format(),
                        verbosity,
                        _MOE2017_VERBOSITY_LEVEL,
                    )

                    # Add the probability
                    q3_prob = Moecache[
                        "rinterpolator_q_given_{}_log10{}".format(m_label, p_label)
                    ].interpolate([secondary_mass / primary_mass])[0]
                    prob_dict["q3"] = q3_prob
                    verbose_print(
                        "\tMoe_di_Stefano_2017_pdf: appended mass ratio (M_1+M_2={} M_3={} P={} q={}) probability ({}) to the prob list ({}) ".format(
                            options["M_1+M_2"],
                            options["M_3"],
                            options["P"],
                            secondary_mass / primary_mass,
                            prob_dict["q3"],
                            prob_dict,
                        ),
                        verbosity,
                        _MOE2017_VERBOSITY_LEVEL,
                    )
                    # if prob_dict['q3'] == 0: # If the probability is 0 then we don't have to calculate more
                    #     calc_total_probdens(prob_dict)
                    #     return prob_dict

                    # TODO ecc 3

    # check for input of multiplicity
    elif multiplicity not in range(1, 5):
        msg = "\tMoe_di_Stefano_2017_pdf: Unknown multiplicity {}".format(multiplicity)
        verbose_print(
            msg,
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
        raise ValueError(msg)

    # Calculate total probdens:
    prob_dict = calc_total_probdens(prob_dict)

    # Some info
    if multiplicity == 1:
        verbose_print(
            "\tMoe_di_Stefano_2017_pdf: M_1={} q=N/A log10P=N/A ({}): {} -> {}\n".format(
                options["M_1"],
                len(prob_dict),
                str(prob_dict),
                prob_dict["total_probdens"],
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    elif multiplicity == 2:
        verbose_print(
            "\tMoe_di_Stefano_2017_pdf: M_1={} q={} log10P={} ecc={} ({}): {} -> {}\n".format(
                options["M_1"],
                options["M_2"] / options["M_1"] if options.get("M_2", None) else "N/A",
                np.log10(options["P"]),
                options["ecc"] if options.get("ecc", None) else "N/A",
                len(prob_dict),
                str(prob_dict),
                prob_dict["total_probdens"],
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    elif multiplicity == 3:
        verbose_print(
            "\tMoe_di_Stefano_2017_pdf: M_1={} q={} log10P={} ecc={} M_3={} log10P2={} ecc2={} ({}): {} -> {}".format(
                options["M_1"],
                options["M_2"] / options["M_1"] if options.get("M_2", None) else "N/A",
                np.log10(options["P"]),
                options["ecc"] if options.get("ecc", None) else "N/A",
                options["M_3"],
                np.log10(options["P2"]),
                options["ecc2"] if options.get("ecc2", None) else "N/A",
                len(prob_dict),
                str(prob_dict),
                prob_dict["total_probdens"],
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    elif multiplicity == 4:
        verbose_print(
            "Moe_di_Stefano_2017_pdf: M_1={} q={} log10P={} ecc={} M_3={} log10P2={} ecc2={} M_4={} log10P3={} ecc3={} ({}) : {} -> {}".format(
                options["M_1"],
                options["M_2"] / options["M_1"] if options.get("M_2", None) else "N/A",
                np.log10(options["P"]),
                options["ecc"] if options.get("ecc", None) else "N/A",
                options["M_3"],
                np.log10(options["P2"]),
                options["ecc2"] if options.get("ecc2", None) else "N/A",
                options["M_4"],
                np.log10(options["P3"]),
                options["ecc3"] if options.get("ecc3", None) else "N/A",
                len(prob_dict),
                str(prob_dict),
                prob_dict["total_probdens"],
            ),
            verbosity,
            _MOE2017_VERBOSITY_LEVEL,
        )
    return prob_dict



"""
Unittests for spacing_functions module
"""

import unittest
import numpy as np

from binarycpython.utils.spacing_functions import const
from binarycpython.utils.functions import Capturing


class test_spacing_functions(unittest.TestCase):
    """
    Unit test for spacing functions
    """

    def test_const(self):
        with Capturing() as output:
            self._test_const()

    def _test_const(self):
        """
        Unittest for function const
        """

        const_return = const(1, 10, 10)
        self.assertTrue(
            (const_return == np.linspace(1, 10, 10 + 1)).all(),
            msg="Output didn't contain SINGLE_STAR_LIFETIME",
        )


if __name__ == "__main__":
    unittest.main()



"""
Module containing functions to interact with binary_c's
population ensemble using the binarycpython package

"""

import sys
import bz2
import time
import json
from typing import Any
import gc
import gzip
import inspect


from halo import Halo
import msgpack
import py_rinterpolate

import simplejson

# import orjson

from binarycpython.utils.dicts import (
    keys_to_floats,
    recursive_change_key_to_float,
    custom_sort_dict,
    recursive_change_key_to_string,
)
from binarycpython.utils.functions import verbose_print


def ensemble_setting(ensemble, parameter_name):
    """
    Function to get the setting of parameter_name in the given ensemble, or return the default value.
    """
    value = None

    try:
        value = ensemble["metadata"]["settings"]["population_settings"]["bse_options"][
            parameter_name
        ]
    except KeyError:
        value = None

    if value is None:
        try:
            value = ensemble["metadata"]["settings"]["population_settings"][
                "grid_options"
            ][parameter_name]
        except KeyError:
            value = None

    if value is None:
        try:
            value = ensemble["metadata"]["settings"]["population_settings"][
                "custom_options"
            ][parameter_name]
        except KeyError:
            value = None

    # not found, try the default
    if value is None:
        try:
            value = ensemble["metadata"]["settings"]["binary_c_defaults"][
                parameter_name
            ]
        except KeyError:
            value = None

    return value


def open_ensemble(filename):
    """
    Function to open an ensemble at filename for reading and decompression if required.
    """
    compression = ensemble_compression(filename)
    if ensemble_file_type(filename) is "msgpack":
        flags = "rb"
    else:
        flags = "rt"
    if compression is "bzip2":
        file_object = bz2.open(filename, flags)
    elif compression is "gzip":
        file_object = gzip.open(filename, flags)
    else:
        file_object = open(filename, flags)
    return file_object


def ensemble_compression(filename):
    """
    Return the compression type of the ensemble file, based on its filename extension.
    """
    if filename.endswith(".bz2"):
        return "bzip2"
    elif filename.endswith(".gz"):
        return "gzip"
    else:
        return None


def ensemble_file_type(filename):
    """
    Returns the file type of an ensemble file.
    """
    if ".json" in filename:
        filetype = "JSON"
    elif ".msgpack" in filename:
        filetype = "msgpack"
    else:
        filetype = None
    return filetype


def load_ensemble(filename, convert_float_keys=True, select_keys=None, timing=False):
    """
    Function to load an ensemeble file, even if it is compressed,
    and return its contents to as a Python dictionary.

    Args:
        convert_float_keys : if True, converts strings to floats.
        select_keys : a list of keys to be selected from the ensemble.
    """

    # open the file

    # load with some info to the terminal
    print("Loading JSON...")

    # open the ensemble and get the file type
    file_object = open_ensemble(filename)
    filetype = ensemble_file_type(filename)

    if not filetype or not file_object:
        print(
            "Unknown filetype : your ensemble should be saved either as JSON or msgpack data."
        )
        sys.exit()

    with Halo(text="Loading", interval=250, spinner="moon", color="yellow"):
        tstart = time.time()
        _loaded = False

        def _hook(obj):
            nonlocal _loaded
            if _loaded == False:
                _loaded = True
                print(
                    "\nLoaded {} data, now putting in a dictionary".format(filetype),
                    flush=True,
                )
            return obj

        if filetype is "JSON":
            # orjson promises to be fast, but it doesn't seem to be
            # and fails on "Infinity"... oops
            # data = orjson.loads(file_object.read())

            # simplejson is faster than standard json and "just works"
            # on the big Moe set in 37s
            data = simplejson.load(file_object, object_hook=_hook)

            # standard json module
            # on the big Moe set takes 42s
            # data = json.load(file_object,
            #                 object_hook=_hook)
        elif filetype is "msgpack":
            data = msgpack.load(file_object, object_hook=_hook)

        if timing:
            print(
                "\n\nTook {} s to load the data\n\n".format(time.time() - tstart),
                flush=True,
            )

    # strip non-selected keys, if a list is given in select_keys
    if select_keys:
        keys = list(data["ensemble"].keys())
        for key in keys:
            if not key in select_keys:
                del data["ensemble"][key]

    # perhaps convert floats?
    tstart = time.time()
    if convert_float_keys:
        # timings are for 100 iterations on the big Moe data set
        # data = format_ensemble_results(data) # 213s
        # data = recursive_change_key_to_float(data) # 61s
        data = keys_to_floats(data)  # 6.94s

    if timing:
        print(
            "\n\nTook {} s to convert floats\n\n".format(time.time() - tstart),
            flush=True,
        )

    # return data
    return data


#####
def extract_ensemble_json_from_string(binary_c_output: str) -> dict:
    """
    Function to extract the ensemble_json information from a raw binary_c output string

    Args:
        binary_c_output: raw binary_c output string

    Returns:
        JSON dictionary with the parsed ENSEMBLE_JSON data
    """

    json_dict = None

    try:
        # If there is no output just return an empty dict:
        if not binary_c_output:
            json_dict = {}
            return json_dict

        ensemble_jsons_strings = [
            line
            for line in binary_c_output.splitlines()
            if line.startswith("ENSEMBLE_JSON")
        ]

        json_dict = handle_ensemble_string_to_json(
            ensemble_jsons_strings[0][len("ENSEMBLE_JSON ") :]
        )

        if len(ensemble_jsons_strings) > 1:
            verbose_print(
                "Warning: There is more than one line starting with ENSEMBLE_JSON. Taking the first, but you should check this out.",
                1,
                0,
            )
    except IndexError:
        verbose_print(
            "Error: Couldn't extract the ensemble information from the output string",
            0,
            0,
        )

    return json_dict


def handle_ensemble_string_to_json(raw_output):
    """
    Function that deals with the raw output of the ensemble and
    creates a working JSON dictionary out of it.

    Having this wrapper makes it easy to

    Args:
        raw_output: raw output of the ensemble dump by binary_c

    Returns:
        json.loads(raw_output, cls=binarycDecoder)

    """

    # return json.loads(json.dumps(ast.literal_eval(raw_output)), cls=binarycDecoder)
    return json.loads(raw_output, cls=binarycDecoder)


def binaryc_json_serializer(obj: Any) -> Any:
    """
    Custom serialiser for binary_c to use when functions are present in the dictionary
    that we want to export.

    Function objects will be turned into str representations of themselves

    Args:
        obj: The object that might not be serialisable

    Returns:
        Either string representation of object if the object is a function, or the object itself
    """

    if inspect.isfunction(obj) or isinstance(obj, py_rinterpolate.Rinterpolate):
        return str(obj)
    return obj


class binarycDecoder(json.JSONDecoder):
    """
    Custom decoder to transform the numbers that are strings to actual floats
    """

    def decode(self, s):
        """
        Entry point function for decoding
        """

        result = super().decode(
            s
        )  # result = super(Decoder, self).decode(s) for Python 2.x
        return self._decode(result)

    def _decode(self, o):
        """
        Depending on the type of object is will determine whether to loop over the elements,
        or try to change the type of the object from string to float

        The try except might be a somewhat rough solution but it catches all cases.
        """

        # Check if we can turn it into a float
        # if isinstance(o, str) or isinstance(o, unicode):
        if isinstance(o, str):
            try:
                return float(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return {k: self._decode(v) for k, v in o.items()}
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o


class BinaryCEncoder(json.JSONEncoder):
    """
    Encoding class function to attempt to convert things to strings.
    """

    def default(self, o):
        """
        Converting function. Well, could be more precise. look at the JSON module
        """
        try:
            str_repr = str(o)
        except TypeError:
            pass
        else:
            return str_repr

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)


def format_ensemble_results(ensemble_dictionary):
    """
    Function to handle all the steps of formatting the ensemble output again.

    Input:
        ensemble_dictionary: dictionary containing all the ensemble results
    """

    original_ensemble_results = ensemble_dictionary

    float_format_ensemble_results = recursive_change_key_to_float(
        original_ensemble_results
    )
    del original_ensemble_results
    gc.collect()

    # Then sort the dictionary
    sorted_ensemble_results = custom_sort_dict(float_format_ensemble_results)
    del float_format_ensemble_results
    gc.collect()

    # Then Change the keys back to a string but with a %g format.
    reformatted_ensemble_results = recursive_change_key_to_string(
        sorted_ensemble_results
    )
    del sorted_ensemble_results
    gc.collect()

    # Put back in the dictionary
    return reformatted_ensemble_results



"""
Unittests for the c-bindings
"""

import os
import sys
import time
import json
import textwrap
import unittest
import numpy as np

from binarycpython import _binary_c_bindings

from binarycpython.utils.functions import (
    temp_dir,
    verbose_print,
    is_capsule,
    Capturing,
)
from binarycpython.utils.ensemble import (
    binarycDecoder,
    handle_ensemble_string_to_json,
    extract_ensemble_json_from_string,
)
from binarycpython.utils.dicts import (
    inspect_dict,
    merge_dicts,
)


TMP_DIR = temp_dir("tests", "test_c_bindings")

#### some useful functions
def return_argstring(
    m1=15.0,
    m2=14.0,
    separation=0,
    orbital_period=453000000000,
    eccentricity=0.0,
    metallicity=0.02,
    max_evolution_time=15000,
    defer_ensemble=0,
    ensemble_filters_off=1,
    ensemble_filter="SUPERNOVAE",
):
    """
    Function to make a argstring that we can use in these tests
    """

    # Make the argstrings
    argstring_template = "binary_c M_1 {0:g} M_2 {1:g} separation {2:g} orbital_period {3:g} \
eccentricity {4:g} metallicity {5:g} max_evolution_time {6:g} ensemble 1 ensemble_defer {7} \
ensemble_filters_off {8} ensemble_filter_{9} 1 probability 0.1 ensemble_dt 1000"

    argstring = argstring_template.format(
        m1,
        m2,
        separation,
        orbital_period,
        eccentricity,
        metallicity,
        max_evolution_time,
        defer_ensemble,
        ensemble_filters_off,
        ensemble_filter,
    )

    return argstring


#######################################################################################################################################################
### General run_system test
#######################################################################################################################################################


class test_run_system(unittest.TestCase):
    """
    Unit test for run_system
    """

    def test_output(self):
        with Capturing() as output:
            self._test_output()

    def _test_output(self):
        """
        General test if run_system works
        """
        print(self.id())

        m1 = 15.0  # Msun
        m2 = 14.0  # Msun
        separation = 0  # 0 = ignored, use period
        orbital_period = 4530.0  # days
        eccentricity = 0.0
        metallicity = 0.02
        max_evolution_time = 15000
        argstring = "binary_c M_1 {0:g} M_2 {1:g} separation {2:g} orbital_period {3:g} eccentricity {4:g} metallicity {5:g} max_evolution_time {6:g}  ".format(
            m1,
            m2,
            separation,
            orbital_period,
            eccentricity,
            metallicity,
            max_evolution_time,
        )

        output = _binary_c_bindings.run_system(argstring=argstring)

        self.assertIn(
            "SINGLE_STAR_LIFETIME",
            output,
            msg="Output didn't contain SINGLE_STAR_LIFETIME",
        )


#######################################################################################################################################################
### memaddr test
#######################################################################################################################################################

# TODO: Make some assertion tests in c


class test_return_store_memaddr(unittest.TestCase):
    """
    Unit test for return_store_memaddr
    """

    def test_return_store_memaddr(self):
        with Capturing() as output:
            self._test_return_store_memaddr()

    def _test_return_store_memaddr(self):
        """
        Test to see if the memory adress is returned properly
        """

        output = _binary_c_bindings.return_store_memaddr()

        # print("function: test_return_store")
        # print("store memory adress:")
        # print(textwrap.indent(str(output), "\t"))

        self.assertTrue(is_capsule(output))
        # self.assertNotEqual(
        #     output, 0, "memory adress seems not to have a correct value"
        # )

        # TODO: check if we can built in some signal for how successful this was.
        _ = _binary_c_bindings.free_store_memaddr(output)


#######################################################################################################################################################
### ensemble tests
#######################################################################################################################################################


class TestEnsemble(unittest.TestCase):
    """
    Unittests for handling the ensemble outputs and adding those
    """

    def __init__(self, *args, **kwargs):
        """
        init function
        """
        super(TestEnsemble, self).__init__(*args, **kwargs)

    def test_return_persistent_data_memaddr(self):
        with Capturing() as output:
            self._test_return_persistent_data_memaddr()

    def _test_return_persistent_data_memaddr(self):
        """
        Test case to check if the memory adress has been created succesfully
        """
        print(self.id())

        output = _binary_c_bindings.return_persistent_data_memaddr()

        self.assertTrue(is_capsule(output), msg="Object must be a capsule")
        # self.assertNotEqual(
        #     output, 0, "memory adress seems not to have a correct value"
        # )

    def test_minimal_ensemble_output(self):
        with Capturing() as output:
            self._test_minimal_ensemble_output()

    def _test_minimal_ensemble_output(self):
        """
        test_case to check if the ensemble output is correctly output
        """
        print(self.id())

        m1 = 2  # Msun
        m2 = 0.1  # Msun

        # Direct output commands
        argstring_1 = return_argstring(
            m1=m1,
            m2=m2,
            ensemble_filter="STELLAR_TYPE_COUNTS",
            defer_ensemble=0,  # no defer to memory location. just output it
        )

        output_1 = _binary_c_bindings.run_system(argstring=argstring_1)

        # Check if the ENSEMBLE_JSON is uberhaubt in the output
        self.assertIn("ENSEMBLE_JSON", output_1)

        test_json = extract_ensemble_json_from_string(output_1)
        self.assertIn("number_counts", test_json)
        self.assertNotEqual(test_json["number_counts"], {})

    def test_minimal_ensemble_output_defer(self):
        with Capturing() as output:
            self._test_minimal_ensemble_output_defer()

    def _test_minimal_ensemble_output_defer(self):
        """
        test_case to check if the ensemble output is correctly output, by using defer command and freeing+outputting
        """
        print(self.id())

        m1 = 2  # Msun
        m2 = 0.1  # Msun

        persistent_data_memaddr = _binary_c_bindings.return_persistent_data_memaddr()

        # Direct output commands
        argstring_1 = return_argstring(
            m1=m1,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",
            defer_ensemble=1,  # no defer to memory location. just output it
        )

        output_1 = _binary_c_bindings.run_system(
            argstring=argstring_1, persistent_data_memaddr=persistent_data_memaddr
        )

        #
        self.assertNotIn("ENSEMBLE_JSON", output_1)

        # free memory and output the stuff.
        raw_json_output = (
            _binary_c_bindings.free_persistent_data_memaddr_and_return_json_output(
                persistent_data_memaddr
            )
        )
        ensemble_json_output = extract_ensemble_json_from_string(raw_json_output)

        self.assertIn("number_counts", ensemble_json_output)
        self.assertNotEqual(ensemble_json_output["number_counts"], {})

    def test_add_ensembles_direct(self):
        with Capturing() as output:
            self._test_add_ensembles_direct()

    def _test_add_ensembles_direct(self):
        """
        test_case to check if adding the ensemble outputs works. Many things should be caught by tests in the merge_dict test, but still good to test a bit here
        """
        print(self.id())

        m1 = 2  # Msun
        m2 = 0.1  # Msun

        # Direct output commands
        argstring_1 = return_argstring(
            m1=m1,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",  # no defer to memory location. just output it
        )
        argstring_2 = return_argstring(
            m1=10,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",  # no defer to memory location. just output it
        )

        #
        output_1 = _binary_c_bindings.run_system(argstring=argstring_1)
        output_2 = _binary_c_bindings.run_system(argstring=argstring_2)

        #
        output_json_1 = extract_ensemble_json_from_string(output_1)
        output_json_2 = extract_ensemble_json_from_string(output_2)

        #
        merged_dict = merge_dicts(output_json_1, output_json_2)

        self.assertIn("number_counts", merged_dict)
        self.assertIn("stellar_type", merged_dict["number_counts"])

        for key in output_json_1["number_counts"]["stellar_type"]["0"]:
            self.assertIn(key, merged_dict["number_counts"]["stellar_type"]["0"])

        for key in output_json_2["number_counts"]["stellar_type"]["0"]:
            self.assertIn(key, merged_dict["number_counts"]["stellar_type"]["0"])

        # compare stuff:
        self.assertLess(
            np.abs(
                output_json_1["number_counts"]["stellar_type"]["0"]["CHeB"]
                + output_json_2["number_counts"]["stellar_type"]["0"]["CHeB"]
                - merged_dict["number_counts"]["stellar_type"]["0"]["CHeB"]
            ),
            1e-10,
        )
        self.assertLess(
            np.abs(
                output_json_1["number_counts"]["stellar_type"]["0"]["MS"]
                + output_json_2["number_counts"]["stellar_type"]["0"]["MS"]
                - merged_dict["number_counts"]["stellar_type"]["0"]["MS"]
            ),
            1e-10,
        )

    def test_compare_added_systems_with_double_deferred_systems(self):
        with Capturing() as output:
            self._test_compare_added_systems_with_double_deferred_systems()

    def _test_compare_added_systems_with_double_deferred_systems(self):
        """
        test to run 2 systems without deferring, and merging them manually. Then run 2 systems with defer and then output them.
        """
        print(self.id())

        m1 = 2  # Msun
        m2 = 0.1  # Msun

        # Direct output commands
        argstring_1 = return_argstring(
            m1=m1,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",  # no defer to memory location. just output it
        )
        argstring_2 = return_argstring(
            m1=10,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",  # no defer to memory location. just output it
        )

        #
        output_1 = _binary_c_bindings.run_system(argstring=argstring_1)
        output_2 = _binary_c_bindings.run_system(argstring=argstring_2)

        #
        output_json_1 = extract_ensemble_json_from_string(output_1)
        output_json_2 = extract_ensemble_json_from_string(output_2)

        #
        merged_dict = merge_dicts(output_json_1, output_json_2)

        ###############################
        # Deferred setup
        persistent_data_memaddr = _binary_c_bindings.return_persistent_data_memaddr()

        argstring_1_deferred = return_argstring(
            m1=m1,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",
            defer_ensemble=1,  # no defer to memory location. just output it
        )
        argstring_2_deferred = return_argstring(
            m1=10,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",
            defer_ensemble=1,  # no defer to memory location. just output it
        )

        # run
        _ = _binary_c_bindings.run_system(
            argstring=argstring_1_deferred,
            persistent_data_memaddr=persistent_data_memaddr,
        )
        _ = _binary_c_bindings.run_system(
            argstring=argstring_2_deferred,
            persistent_data_memaddr=persistent_data_memaddr,
        )

        # output
        raw_json_output = (
            _binary_c_bindings.free_persistent_data_memaddr_and_return_json_output(
                persistent_data_memaddr
            )
        )
        ensemble_json_output = extract_ensemble_json_from_string(raw_json_output)

        # CHeck all keys are present
        for key in merged_dict["number_counts"]["stellar_type"]["0"]:
            self.assertIn(
                key, ensemble_json_output["number_counts"]["stellar_type"]["0"]
            )

        # Check if they are of the same value
        for key in merged_dict["number_counts"]["stellar_type"]["0"]:
            self.assertLess(
                np.abs(
                    merged_dict["number_counts"]["stellar_type"]["0"][key]
                    - ensemble_json_output["number_counts"]["stellar_type"]["0"][key]
                ),
                1e-10,
            )

    def test_combine_with_empty_json(self):
        with Capturing() as output:
            self._test_combine_with_empty_json()

    def _test_combine_with_empty_json(self):
        """
        Test for merging with an empty dict
        """
        print(self.id())

        m1 = 2  # Msun
        m2 = 0.1  # Msun

        argstring_1 = return_argstring(
            m1=m1,
            m2=m2,
            orbital_period=1000000000,
            ensemble_filter="STELLAR_TYPE_COUNTS",  # no defer to memory location. just output it
        )

        output_1 = _binary_c_bindings.run_system(argstring=argstring_1)

        output_json_1 = extract_ensemble_json_from_string(output_1)

        assert_message = "combining output json with empty dict should give same result as initial json"

        self.assertEqual(merge_dicts(output_json_1, {}), output_json_1, assert_message)

    #############

    # def test_full_ensemble_output(self):
    #     with Capturing() as output:
    #         self._test_full_ensemble_output()

    def _test_full_ensemble_output(self):
        """
        Function to just output the whole ensemble
        TODO: put this one back
        """
        print(self.id())

        m1 = 2  # Msun
        m2 = 0.1  # Msun

        argstring_1 = return_argstring(
            m1=m1, m2=m2, orbital_period=1000000000, ensemble_filter=0
        )
        argstring_1 = return_argstring(defer_ensemble=0, ensemble_filters_off=0)

        output_1 = _binary_c_bindings.run_system(argstring=argstring_1)

        #
        output_json_1 = extract_ensemble_json_from_string(output_1)

        keys = json_1.keys()

        # assert statements:
        self.assertIn("number_counts", keys)
        self.assertIn("HRD", keys)
        self.assertIn("HRD(t)", keys)
        self.assertIn("Xyield", keys)
        self.assertIn("distributions", keys)
        self.assertIn("scalars", keys)


#######################################################################################################################################################
### ensemble tests
#######################################################################################################################################################

if __name__ == "__main__":
    unittest.main()
